# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import json
import logging
import re
from typing import Any, List, Dict, Optional, Set, TYPE_CHECKING

from .constants import *
from ..schema.constants import (
    VERSION_0_9,
    VERSION_0_8,
    A2UI_OPEN_TAG,
    A2UI_CLOSE_TAG,
    SURFACE_ID_KEY,
    CATALOG_COMPONENTS_KEY,
)
from ..schema.validator import (
    analyze_topology,
    extract_component_ref_fields,
    extract_component_required_fields,
)
from .response_part import ResponsePart


if TYPE_CHECKING:
  from ..schema.catalog import A2uiCatalog

logger = logging.getLogger(__name__)


# Keys whose string values can be safely auto-closed (healed) if fragmented in the stream.
# Structural or atomic keys (e.g., id, surfaceId, path) are NOT cuttable to prevent
# incorrect parsing or data binding.
CUTTABLE_KEYS = {
    "literalString",
    "valueString",
    "label",
    "hint",
    "caption",
    "altText",
    "text",
}


class A2uiStreamParser:
  """Parses a stream of text for A2UI JSON messages with fine-grained component yielding.

  This class acts as a factory that returns a version-specific parser instance
  (V08 or V09) depending on the catalog version.
  """

  def __new__(cls, catalog: "A2uiCatalog" = None, *args, **kwargs):
    if cls is A2uiStreamParser:
      version = getattr(catalog, "version", None) if catalog else None
      if version == VERSION_0_9:
        from .streaming_v09 import A2uiStreamParserV09

        return A2uiStreamParserV09(catalog=catalog, *args, **kwargs)
      else:
        from .streaming_v08 import A2uiStreamParserV08

        return A2uiStreamParserV08(catalog=catalog, *args, **kwargs)
    return super().__new__(cls)

  def __init__(self, catalog: "A2uiCatalog" = None):
    self._version = getattr(catalog, "version", None) if catalog else None
    self._ref_fields_map = extract_component_ref_fields(catalog) if catalog else {}
    self._required_fields_map = (
        extract_component_required_fields(catalog) if catalog else {}
    )
    from ..schema.validator import A2uiValidator

    self._validator = A2uiValidator(catalog) if catalog else None

    self._found_delimiter = False
    self._buffer = ""
    self._json_buffer = ""
    self._brace_stack: List[Tuple[str, int]] = []
    self._brace_count = 0
    self._in_top_level_list = False
    self._in_string = False
    self._string_escaped = False

    self._seen_components: Dict[str, Dict[str, Any]] = {}

    # Track data model for path resolution
    self._yielded_data_model: Dict[str, Any] = {}
    self._deleted_surfaces: Set[str] = set()

    # Set of unique component IDs yielded per surface to prevent duplicate yielding
    # surfaceId -> set of cids
    self._yielded_ids: Dict[str, Set[str]] = {}
    # (surfaceId, cid) -> hash of content for change detection
    self._yielded_contents: Dict[Any, str] = {}

    self._root_ids: Dict[str, str] = {}  # The root component IDs mapped per surface
    self._default_root_id: Optional[str] = None  # Base default root ID for the protocol
    self._unbound_root_id: Optional[str] = (
        None  # Temporary holding variable for when root arrives before surfaceId
    )
    self._surface_id: Optional[str] = None  # The active surface ID tracking the context
    self._msg_types: List[str] = []  # Running list of message types seen in the block

    # A set of surface ids for which we have already yielded a start message
    # Tracks if beginRendering or createSurface was emitted
    self._yielded_start_messages: Set[str] = set()

    # The current active message type for component grouping
    self._active_msg_type: Optional[str] = None

    # State for buffering updates until surface is ready
    self._pending_messages: Dict[str, List[Dict[str, Any]]] = (
        {}
    )  # surfaceId -> list of msgs delayed until start message arrives
    self._buffered_start_message: Optional[Dict[str, Any]] = (
        None  # The start message to yield before any components
    )
    self._topology_dirty = False  # Set to true if components are added out of order
    self._in_top_level_list = False
    self._found_valid_json_in_block = False

  @property
  def _placeholder_component(self) -> Dict[str, Any]:
    """Returns the version-specific placeholder component.

    This is used when a component references a child component that hasn't yet
    streamed in. The placeholder component is added to the components list and
    the reference is updated to point to the placeholder component.
    """
    raise NotImplementedError("Subclasses must implement _placeholder_component")

  @property
  def surface_id(self) -> Optional[str]:
    return self._surface_id

  @surface_id.setter
  def surface_id(self, value: Optional[str]):
    self._surface_id = value
    if value is not None and self._unbound_root_id is not None:
      self._root_ids[value] = self._unbound_root_id
      self._unbound_root_id = None

  @property
  def root_id(self) -> Optional[str]:
    if self._surface_id:
      return self._root_ids.get(self._surface_id, self._default_root_id)
    # Return unbound root ID if explicitly sniffed, otherwise use protocol default
    return (
        self._unbound_root_id
        if self._unbound_root_id is not None
        else self._default_root_id
    )

  @root_id.setter
  def root_id(self, value: Optional[str]):
    if self._surface_id:
      if value is not None:
        self._root_ids[self._surface_id] = value
      else:
        self._root_ids.pop(self._surface_id, None)
    else:
      self._unbound_root_id = value

  @property
  def msg_types(self) -> List[str]:
    return self._msg_types

  def add_msg_type(self, msg_type: str):
    if msg_type not in self._msg_types:
      self._msg_types.append(msg_type)
    if msg_type in (
        MSG_TYPE_SURFACE_UPDATE,
        MSG_TYPE_UPDATE_COMPONENTS,
        MSG_TYPE_CREATE_SURFACE,
    ):
      self._active_msg_type = msg_type

  @property
  def _yielded_surfaces_set(self) -> Set[str]:
    """Provides access to version-specific yielded surfaces set."""
    raise NotImplementedError("Subclasses must implement _yielded_surfaces_set")

  def is_protocol_msg(self, obj: Dict[str, Any]) -> bool:
    """Checks if the object is a recognized A2UI message for this version."""
    raise NotImplementedError("Subclasses must implement is_protocol_msg")

  @property
  def _data_model_msg_type(self) -> str:
    """Returns the message type identifier for data model updates."""
    raise NotImplementedError("Subclasses must implement _data_model_msg_type")

  def _get_active_msg_type_for_components(self) -> Optional[str]:
    """Determines which msg_type to use when wrapping component updates."""
    raise NotImplementedError(
        "Subclasses must implement _get_active_msg_type_for_components"
    )

  def _deduplicate_data_model(self, m: Dict[str, Any], strict_integrity: bool) -> bool:
    """Returns True if message should be yielded, False if skipped."""
    return True

  def _yield_messages(
      self,
      messages_to_yield: List[Dict[str, Any]],
      messages: List[ResponsePart],
      strict_integrity: bool = True,
  ):
    """Validates and appends messages to the final output list."""
    for m in messages_to_yield:
      if not self._deduplicate_data_model(m, strict_integrity):
        continue

      # Each surface update message must specify a surfaceId and satisfy catalog validation.
      if self._validator:
        try:
          self._validator.validate(
              m, root_id=self.root_id, strict_integrity=strict_integrity
          )
        except ValueError as e:
          if strict_integrity:
            raise e
          else:
            logger.debug(f"Validation failed for partial/sniffed message: {e}")
            continue

      # Consolidated appending logic
      if messages and messages[-1].a2ui_json is None:
        messages[-1].a2ui_json = [m]
      elif messages and isinstance(messages[-1].a2ui_json, list):
        messages[-1].a2ui_json.append(m)
      else:
        messages.append(ResponsePart(a2ui_json=[m]))

  def _delete_surface(self, sid: str) -> None:
    """Clears all state related to a specific surface."""
    self._pending_messages.pop(sid, None)
    self._yielded_ids.pop(sid, None)

    # Clear contents for this surface
    self._yielded_contents = {
        k: v for k, v in self._yielded_contents.items() if k[0] != sid
    }
    self._yielded_surfaces_set.discard(sid)
    self._yielded_start_messages.discard(sid)

    self._deleted_surfaces.add(sid)

  def process_chunk(self, chunk: str) -> List[ResponsePart]:
    """Processes a chunk of text and returns any complete A2UI messages found.

    This is the primary entry point for the streaming parser. It handles the
    initial "tag hunt" and then delegates JSON fragment processing to
    `_process_json_chunk`. It supports multiple A2UI blocks in a single stream.

    Args:
        chunk: The chunk of raw text (e.g., from an LLM stream) to process.

    Returns:
        A list of parsed A2UI message dictionaries.
    """
    messages = []
    self._buffer += chunk

    while True:
      if not self._found_delimiter:
        # Looking for <a2ui-json>
        if A2UI_OPEN_TAG in self._buffer:
          parts = self._buffer.split(A2UI_OPEN_TAG, 1)
          if parts[0]:
            messages.append(ResponsePart(text=parts[0]))
          self._found_delimiter = True
          self._buffer = parts[1]
          # Continue to process the content after the open tag
        else:
          # Yield conversational text while avoiding split tags
          keep_len = 0
          for i in range(len(A2UI_OPEN_TAG) - 1, 0, -1):
            if self._buffer.endswith(A2UI_OPEN_TAG[:i]):
              keep_len = i
              break

          if len(self._buffer) > keep_len:
            safe_to_yield = len(self._buffer) - keep_len
            text_to_yield = self._buffer[:safe_to_yield]
            messages.append(ResponsePart(text=text_to_yield))
            self._buffer = self._buffer[safe_to_yield:]
          break

      if self._found_delimiter:
        # Looking for </a2ui-json>
        if A2UI_CLOSE_TAG in self._buffer:
          parts = self._buffer.split(A2UI_CLOSE_TAG, 1)
          json_fragment = parts[0]
          self._process_json_chunk(json_fragment, messages)
          if not self._found_valid_json_in_block:
            raise ValueError(
                "Failed to parse JSON: No valid JSON object found in A2UI block."
            )

          # End of block: reset JSON state but keep seen_components
          self._found_delimiter = False
          self._reset_json_state()

          self._buffer = parts[1]
          # Continue loop to look for next A2UI_OPEN_TAG in remaining buffer
        else:
          # Find if the buffer ends with a prefix of A2UI_CLOSE_TAG
          # To avoid split-tag issues, we only delay processing if it looks like a close tag is starting
          keep_len = 0
          for i in range(1, len(A2UI_CLOSE_TAG)):
            if self._buffer.endswith(A2UI_CLOSE_TAG[:i]):
              keep_len = i

          if keep_len < len(self._buffer):
            to_process = self._buffer[: len(self._buffer) - keep_len]
            self._buffer = self._buffer[len(self._buffer) - keep_len :]
            self._process_json_chunk(to_process, messages)
          break

    # Deduplicate surfaceUpdate messages to avoid over-yielding in a single chunk
    for part in messages:
      if not part.a2ui_json:
        continue

      deduped_msgs = []
      seen_su = set()
      # Iterate backwards to keep only the last (most complete) surfaceUpdate for each surface
      for m in reversed(part.a2ui_json):
        is_su = False
        sid = None
        if isinstance(m, dict) and MSG_TYPE_SURFACE_UPDATE in m:
          is_su = True
          sid = m[MSG_TYPE_SURFACE_UPDATE].get(SURFACE_ID_KEY)

        if is_su and sid:
          if sid not in seen_su:
            deduped_msgs.append(m)
            seen_su.add(sid)
        else:
          deduped_msgs.append(m)

      deduped_msgs.reverse()
      part.a2ui_json = deduped_msgs

    if messages:
      logger.debug(
          f"DEBUG: process_chunk returning {len(messages)} messages: {messages}"
      )
    return messages

  def _reset_json_state(self):
    """Resets the JSON-specific parsing state (e.g., at the end of a block)."""
    self._json_buffer = ""
    self._brace_stack = []
    self._brace_count = 0
    self._in_top_level_list = False
    self._in_string = False
    self._string_escaped = False
    self._msg_types = []
    self._found_valid_json_in_block = False
    # Note: we do NOT reset _active_msg_type or _yielded_contents here

    # so re-yielding works between blocks

  def _fix_json(self, fragment: str) -> str:
    """Attempts to fix a partial JSON fragment by adding missing closing delimiters."""
    fixed = fragment.rstrip()
    if not fixed:
      return ""

    stack = []
    in_string = False
    escaped = False
    last_quote_idx = -1

    # Single pass to track strings and braces
    for i, char in enumerate(fixed):
      if escaped:
        escaped = False
        continue
      if char == "\\":
        escaped = True
        continue
      if char == '"':
        in_string = not in_string
        if in_string:
          last_quote_idx = i
      elif not in_string:
        if char in ("{", "["):
          stack.append(char)
        elif char in ("}", "]"):
          if stack:
            stack.pop()

    # 1. Close open strings (healing)
    if in_string:
      # We only auto-close strings for safe keys (CUTTABLE_KEYS)
      prefix = fixed[:last_quote_idx].rstrip()
      if prefix.endswith(":"):
        key_match = re.findall(r'"([^"]+)"\s*:\s*$', prefix)
        if key_match:
          key = key_match[0]
          if key not in CUTTABLE_KEYS:
            return ""

          # Special case: don't cut URL bindings, as partial URLs break images/links
          if key == "valueString":
            string_val = fixed[last_quote_idx + 1 :]
            if (
                string_val.startswith("http://")
                or string_val.startswith("https://")
                or string_val.startswith("data:")
                or string_val.startswith("/")
            ):
              return ""

            # Check if this value belongs to a URL-like key in the data model
            # Look backwards in the prefix (max 200 chars) for the "key" assignment
            prev_key_matches = re.findall(r'"key"\s*:\s*"([^"]+)"', prefix[-200:])
            if prev_key_matches:
              data_key = prev_key_matches[-1].lower()
              if any(k in data_key for k in ("url", "link", "src", "href", "image")):
                return ""
      fixed += '"'

    # 2. Clean up trailing comma
    fixed = fixed.rstrip()
    if fixed.endswith(","):
      fixed = fixed[:-1].rstrip()

    # 3. Close braces and brackets
    while stack:
      opening = stack.pop()
      fixed += "}" if opening == "{" else "]"

    return fixed

  def _process_json_chunk(self, chunk: str, messages: List[ResponsePart]):
    for char in chunk:
      char_handled = False
      if not self._in_top_level_list:
        if char == "[":
          if self._brace_count == 0:
            self._in_top_level_list = True
          self._brace_stack.append(("[", len(self._json_buffer)))
          self._json_buffer += "["
          self._brace_count += 1
          char_handled = True
        else:
          continue

      # Track string state to avoid miscounting braces inside strings
      if not char_handled and self._in_string:
        if self._string_escaped:
          self._string_escaped = False
          if self._brace_count > 0:
            self._json_buffer += char
        elif char == "\\":
          self._string_escaped = True
          if self._brace_count > 0:
            self._json_buffer += char
        elif char == '"':
          self._in_string = False
          if self._brace_count > 0:
            self._json_buffer += char
        else:
          if self._brace_count > 0:
            self._json_buffer += char
        char_handled = True

      if not char_handled:
        if char == '"':
          self._in_string = True
          self._string_escaped = False
          if self._brace_count > 0:
            self._json_buffer += char
        elif char == "{":
          if self._brace_count == 0:
            self._msg_types = []
          # Store (type, index) on stack
          self._brace_stack.append(("{", len(self._json_buffer)))
          self._json_buffer += "{"
          self._brace_count += 1
        elif char == "}":
          # Trigger object recognition
          # In v0.8 streaming, we might be nested inside surfaceUpdate/components list
          # So we check if it looks like a component even if brace_count > 1
          if self._brace_stack:  # Ensure there's an opening brace to pop
            # Pop the typed entry
            b_type, start_idx = self._brace_stack.pop()
            # If we popped a bracket while looking for a brace, we have a mismatch
            # but we'll be resilient and just continue
            self._json_buffer += "}"
            self._brace_count -= 1

            if self._brace_count >= 0:  # Allow processing even if not top-level object
              # The `i` here is the index within the current `chunk`.
              # We need to get the full buffer from `start_idx` in `_json_buffer`
              # up to the current point where `char` (which is '}') was added.
              # The `_json_buffer` already has `char` appended.
              obj_buffer = self._json_buffer[start_idx:]
              if obj_buffer.startswith("{") and obj_buffer.endswith("}"):
                try:
                  obj = json.loads(obj_buffer)
                  if isinstance(obj, dict):
                    self._found_valid_json_in_block = True
                    logger.debug(
                        f"[Parsed Dict] Keys: {list(obj.keys())}, protocol check"
                        " follows..."
                    )

                    is_protocol = self._in_top_level_list and self.is_protocol_msg(obj)
                    is_comp = obj.get("id") and obj.get("component")
                    # Process objects at top-level OR items in top-level list
                    # When in a list, we are top-level if the ONLY thing on the stack is the list opener
                    is_top_level = (len(self._brace_stack) == 0) or (
                        self._in_top_level_list
                        and len(self._brace_stack) == 1
                        and self._brace_stack[0][0] == "["
                    )
                    if is_comp:
                      self._handle_partial_component(obj, messages)
                    elif is_top_level or is_protocol:
                      if not self._handle_complete_object(
                          obj, self.surface_id, messages
                      ):
                        # Not a recognized message type. Validate to catch schema errors.
                        self._yield_messages([obj], messages)

                    if self._brace_count == 0 or (
                        self._in_top_level_list and len(self._brace_stack) == 1
                    ):
                      # Aggressively clear processed objects from the buffer to prevent slowdown.
                      if len(self._brace_stack) == 1 and self._brace_stack[0][0] == "[":
                        # Keep '[' and remove the object after it
                        self._json_buffer = (
                            self._json_buffer[:start_idx]
                            + self._json_buffer[start_idx + len(obj_buffer) :]
                        )
                      else:
                        self._json_buffer = self._json_buffer[len(obj_buffer) :]
                        if self._brace_stack:
                          shift = len(obj_buffer)
                          self._brace_stack = [
                              (b_t, i - shift) for b_t, i in self._brace_stack
                          ]

                except json.JSONDecodeError as e:
                  logger.debug(f"Object recognition failed: {e}")

        elif char == "[":
          self._brace_stack.append(("[", len(self._json_buffer)))
          self._json_buffer += "["
          self._brace_count += 1
        elif char == "]":
          if self._brace_stack and self._brace_stack[-1][0] == "[":
            # Pop the typed entry
            b_type, start_idx = self._brace_stack.pop()
            self._json_buffer += "]"
            self._brace_count -= 1
            if self._brace_count == 0:
              self._in_top_level_list = False
        else:
          if self._brace_count > 0:
            self._json_buffer += char

      # Sniff for metadata reactively on key delimiters to catch identifiers early
      if self._brace_count > 0 and char in ('"', ":", ",", "}", "]"):
        self._sniff_metadata()

    # Sniff for partial components at the end of the chunk
    if self._brace_count >= 1 and self._json_buffer:
      self._sniff_partial_component(messages)
      self._sniff_partial_data_model(messages)

    if self._topology_dirty:
      self.yield_reachable(messages, check_root=False, raise_on_orphans=False)
      self._topology_dirty = False

  def _construct_sniffed_data_model_message(
      self, active_msg_type: str, delta_msg_payload: Dict[str, Any]
  ) -> Dict[str, Any]:
    """Returns the message to yield for a partial data model update."""
    return {active_msg_type: delta_msg_payload}

  def _sniff_partial_data_model(self, messages: List[ResponsePart]) -> None:
    msg_type = self._data_model_msg_type
    if f'"{msg_type}"' not in self._json_buffer:
      return
    # Look through the brace stack for objects that might contain data model updates
    for b_type, start_idx in reversed(self._brace_stack):
      if b_type != "{":
        continue
      raw_fragment = self._json_buffer[start_idx:]
      if not raw_fragment:
        continue

      fixed_fragment = self._fix_json(raw_fragment)
      obj = None
      try:
        obj = json.loads(fixed_fragment)
      except json.JSONDecodeError:
        # Fallback: iteratively strip from the last comma
        # This handles cases where _fix_json produces invalid JSON
        # from an incomplete trailing element (e.g. `{"key"}` from `{"ke`)
        trimmed = raw_fragment
        while "," in trimmed:
          trimmed = trimmed.rsplit(",", 1)[0]
          try:
            fixed_trimmed = self._fix_json(trimmed)
            if fixed_trimmed:
              obj = json.loads(fixed_trimmed)
              break
          except json.JSONDecodeError:
            continue

      if obj and isinstance(obj, dict):
        active_msg_type = None
        msg_type = self._data_model_msg_type
        if msg_type in obj:
          active_msg_type = msg_type

        if active_msg_type:
          dm_obj = obj[active_msg_type]
          if isinstance(dm_obj, dict) and "contents" in dm_obj:

            raw_contents = dm_obj["contents"]
            contents_dict = self._parse_contents_to_dict(raw_contents)

            if contents_dict:
              delta = {}
              for k, v in contents_dict.items():
                if self._yielded_data_model.get(k) != v:
                  delta[k] = v

              if delta:
                sid = dm_obj.get(SURFACE_ID_KEY) or self._surface_id or "default"
                # Deduplicate delta_contents by only keeping the LATEST entry for each dirty key
                if isinstance(raw_contents, list):
                  delta_contents = []
                  seen_keys = set()
                  for entry in reversed(raw_contents):
                    if not isinstance(entry, dict):
                      continue
                    k = entry.get("key")
                    # Only include entries that have a valid parsed key (cumulative)
                    if k and k in contents_dict and k not in seen_keys:
                      delta_contents.insert(0, entry)
                      seen_keys.add(k)
                  delta_contents = self._prune_incomplete_datamodel_entries(
                      delta_contents
                  )
                else:
                  delta_contents = delta

                delta_msg_payload = {
                    SURFACE_ID_KEY: sid,
                    "contents": delta_contents,
                }
                if "path" in dm_obj:
                  delta_msg_payload["path"] = dm_obj["path"]

                delta_msg = self._construct_sniffed_data_model_message(
                    active_msg_type, delta_msg_payload
                )
                self._yield_messages([delta_msg], messages, strict_integrity=False)

                self._yielded_data_model.update(contents_dict)
                # Update internal model for path resolution
                self.update_data_model(dm_obj, messages)

  def _sniff_partial_component(self, messages: List[ResponsePart]):
    """Attempts to parse a partial component from the current buffer."""
    # We only care about components if we are inside a "components" array
    if f'"{CATALOG_COMPONENTS_KEY}"' not in self._json_buffer:
      return
    # Try parsing from inner to outer to find the smallest complete component
    for b_type, start_idx in reversed(self._brace_stack):
      if b_type != "{":
        continue
      raw_fragment = self._json_buffer[start_idx:]
      if not raw_fragment:
        continue

      fixed_fragment = self._fix_json(raw_fragment)
      try:
        obj = json.loads(fixed_fragment)
        if isinstance(obj, dict) and obj.get("id") and obj.get("component"):
          if isinstance(obj["component"], str):
            # Flat style (v0.9+): component type is a string
            self._handle_partial_component(obj, messages)
          elif isinstance(obj["component"], dict) and len(obj["component"]) > 0:
            # Structured style (v0.8): Ignore components that are effectively empty (no type keys)
            self._handle_partial_component(obj, messages)

      except Exception:
        continue

  def _sniff_metadata(self) -> None:
    """Sniffs for surfaceId, root, and msg_types in the current json_buffer."""
    raise NotImplementedError("Subclasses must implement _sniff_metadata")

  def _prune_incomplete_datamodel_entries(self, entries: Any) -> Any:
    """Recursively removes data model entries that only contain 'key' and no valid values."""
    if not isinstance(entries, list):
      return entries

    pruned = []
    for entry in entries:
      if not isinstance(entry, dict):
        pruned.append(entry)
        continue

      has_val = False
      for vkey in ("value", "valueString", "valueNumber", "valueBoolean"):
        if vkey in entry:
          has_val = True
          break

      if "valueMap" in entry:
        pruned_map = self._prune_incomplete_datamodel_entries(entry["valueMap"])
        # valueMap is considered valid even if empty, meaning map was explicitly empty
        if isinstance(pruned_map, list):
          if not pruned_map and len(entry["valueMap"]) > 0:
            # If it was non-empty and became empty, it only had incomplete elements. Discard map.
            del entry["valueMap"]
          else:
            entry["valueMap"] = pruned_map
            has_val = True

      if has_val and "key" in entry:
        pruned.append(entry)

    return pruned

  def _handle_partial_component(
      self, comp: Dict[str, Any], messages: List[ResponsePart]
  ):
    """Handles a component discovered before its parent message is finished.

    When the parser sees a full JSON object that looks like a component
    (contains `id` and `component` keys) within a larger message, it caches
    the component and attempts to yield it immediately if it's reachable.

    Args:
        comp: The parsed component dictionary.
        messages: The list to append any renderable partial messages to.
    """
    comp_id = comp.get("id")
    if not comp_id:
      return

    # Skip caching this component if it has empty dictionaries for complex properties.
    # Elements like `children`, `text`, `url`, etc., violate A2UI schema if empty
    # and will crash the client. We want the parent to yield a loading placeholder instead.
    def _has_empty_dict(obj: Any) -> bool:
      if isinstance(obj, dict):
        if not obj:
          return True
        return any(_has_empty_dict(v) for v in obj.values())
      elif isinstance(obj, list):
        return any(_has_empty_dict(v) for v in obj)
      return False

    component_def = comp.get("component")
    if isinstance(component_def, str):
      # v0.9 flat style: check the whole component object for empty dicts
      if _has_empty_dict(comp):
        return
    elif _has_empty_dict(component_def):
      # v0.8 nested style: check properties inside component
      return

    if isinstance(component_def, dict) and hasattr(self, "_required_fields_map"):
      comp_type = next(iter(component_def.keys())) if component_def else None
      if comp_type:
        props = component_def.get(comp_type, {})
        if isinstance(props, dict):
          required_fields = self._required_fields_map.get(comp_type, set())
          for req in required_fields:
            if req not in props:
              return

    self._seen_components[comp_id] = comp
    self._topology_dirty = True

  def _parse_contents_to_dict(self, raw_contents: Any) -> Dict[str, Any]:
    """Recursively parses a list of A2UI contents into a flat dictionary."""
    if isinstance(raw_contents, dict):
      return raw_contents
    if not isinstance(raw_contents, list):
      return {}

    res = {}
    for entry in raw_contents:
      if not isinstance(entry, dict):
        continue
      key = entry.get("key")
      val = None
      for vkey in ["value", "valueString", "valueNumber", "valueBoolean"]:
        if vkey in entry:
          val = entry[vkey]
          break

      if val is None and "valueMap" in entry:
        val = self._parse_contents_to_dict(entry["valueMap"])

      if key and val is not None:
        res[key] = val
    return res

  def update_data_model(
      self, update: Dict[str, Any], messages: List[ResponsePart]
  ) -> None:
    """Updates the internal data model and marks affected components as dirty."""
    # Data model update can be v0.8 flat or v0.9+ contents list
    raw_contents = update.get("contents")
    if raw_contents is not None:
      contents = self._parse_contents_to_dict(raw_contents)
    else:
      # Fallback for old v0.8 flat structure or other layouts
      contents = {
          k: v
          for k, v in update.items()
          if k not in (SURFACE_ID_KEY, "root", "contents")
      }

  def _handle_complete_object(
      self,
      obj: Dict[str, Any],
      sid: Optional[str],
      messages: List[ResponsePart],
  ) -> bool:
    """Handles an object that has been fully parsed. To be implemented by subclasses."""
    raise NotImplementedError("Subclasses must implement _handle_complete_object")

  def yield_reachable(
      self,
      messages: List[Dict[str, Any]],
      check_root: bool = False,
      raise_on_orphans: bool = False,
  ):
    """Yields a partial message containing all reachable and seen components.

    This is the core of the streaming logic. Instead of waiting for a UI message
    which could contain dozens of components, we yield "partial" updates as soon
    as we have enough components to build a renderable sub-tree from the root.

    Args:
        messages: The list to which partial messages will be appended.
        check_root: If True, raises an error if the root component isn't seen yet.
        raise_on_orphans: If True, uses strict topology analysis to catch loops.
    """
    active_msg_type = self._get_active_msg_type_for_components()
    if not self.root_id or not active_msg_type:
      return

    # Buffer components until we have a beginRendering or createSurface for a known surface.
    if not self.surface_id:
      return

    sid = self.surface_id
    if sid not in self._yielded_surfaces_set and not self._buffered_start_message:
      return

    try:
      # Analyze topology of current seen components
      components_to_analyze = list(self._seen_components.values())

      if check_root and self.root_id not in self._seen_components:
        raise ValueError(
            f"No root component (id='{self.root_id}') found in {active_msg_type}"
        )

      reachable_ids = analyze_topology(
          self.root_id,
          components_to_analyze,
          self._ref_fields_map,
          raise_on_orphans=raise_on_orphans,
      )

      # We only yield components we actually have in our "seen" cache
      available_reachable = reachable_ids & set(self._seen_components.keys())

      if check_root and not available_reachable:
        raise ValueError(
            f"No root component (id='{self.root_id}') found in {active_msg_type}"
        )

      # 1. Process placeholders and partial children
      processed_components = []
      extra_components = []
      surface_id = self.surface_id or "unknown"
      yielded_for_surface = self._yielded_ids.get(surface_id, set())

      for rid in sorted(list(available_reachable)):
        comp = copy.deepcopy(self._seen_components[rid])
        # Apply path placeholders and prune unseen children in a single pass
        re_yielding = rid in yielded_for_surface
        self._process_component_topology(
            comp, extra_components, inline_resolved=re_yielding
        )
        processed_components.append(comp)

      # Add generated placeholders to the yield
      processed_components.extend(extra_components)

      # 2. Check if we have NEW or UPDATED reachable components to yield for THIS surface
      surface_id = self.surface_id
      if not surface_id or surface_id in self._deleted_surfaces:
        return

      should_yield = False
      if available_reachable - yielded_for_surface:
        should_yield = True
      else:
        # Check if any yielded component's content has changed for this surface
        for comp in processed_components:
          cid = comp["id"]
          content_str = json.dumps(comp, sort_keys=True)
          state_key = (surface_id, cid)
          if self._yielded_contents.get(state_key) != content_str:
            should_yield = True
            break

      if should_yield:
        current_sid = self.surface_id or "unknown"
        if (
            self._buffered_start_message
            and current_sid not in self._yielded_start_messages
        ):
          self._yield_messages(
              [self._buffered_start_message], messages, strict_integrity=True
          )
          self._yielded_start_messages.add(current_sid)
          self._yielded_surfaces_set.add(current_sid)

        # Construct a partial message of the correct type
        partial_msg = self._construct_partial_message(
            processed_components, active_msg_type
        )

        # Use strict_integrity=False for partial fragments yielded during streaming
        self._yield_messages([partial_msg], messages, strict_integrity=False)
        self._yielded_ids.setdefault(surface_id, set()).update(available_reachable)

        # Update content/placeholder tracking
        for comp in processed_components:
          cid = comp["id"]
          self._yielded_contents[(surface_id, cid)] = json.dumps(comp, sort_keys=True)

    except ValueError as e:
      if "Circular reference detected" in str(e):
        raise e
      # Other topology errors (like orphans) are ignored during streaming
      # as dependencies might still be on the wire.
      msg = str(e)
      if (
          raise_on_orphans
          or "Circular" in msg
          or "Self-reference" in msg
          or "recursion" in msg.lower()
          or check_root
      ):
        logger.debug(f"yield_reachable error (strict={check_root}): {msg}")
        raise e

  def _get_placeholder_id(self, child_id: str) -> str:
    """Returns the ID to use for a missing child placeholder."""
    return f"loading_{child_id}"

  def _process_component_topology(
      self,
      comp: Dict[str, Any],
      extra_components: List[Dict[str, Any]],
      inline_resolved: bool = False,
  ):
    """Recursively processes path placeholders and child pruning in one pass."""
    comp_id = comp.get("id", "unknown")

    # Deduce the component type for better placeholder typing
    comp_type = (
        next(iter(comp.get("component", {}).keys()))
        if comp.get("component") and isinstance(comp.get("component"), dict)
        else ""
    )

    def traverse(obj, parent_key=None):
      if isinstance(obj, dict):
        # 1. Handle Path Placeholders (from _apply_placeholders)
        if (
            "path" in obj
            and isinstance(obj["path"], str)
            and obj["path"].startswith("/")
        ):
          path = obj["path"]
          key = path.lstrip("/")
          if self._version != VERSION_0_9:
            if "componentId" not in obj:
              obj.clear()
            obj.update({"path": "/" + key})
        elif self._version != VERSION_0_9:
          # If not in data model, still ensure path has leading slash if it's a bindable object (v0.8 only)
          current_path = obj.get("path")
          if current_path is not None:
            if not isinstance(current_path, str) or not current_path.startswith("/"):
              obj["path"] = "/" + str(current_path)

        # 2. Handle Child Pruning (from _prune_unseen_children)
        for field in (
            "children",
            "explicitList",
            "child",
            "contentChild",
            "entryPointChild",
            "componentId",
        ):
          if field in obj:
            if isinstance(obj[field], list):
              valid_children = []
              for child_id in obj[field]:
                if child_id in self._seen_components:
                  valid_children.append(child_id)
                else:
                  # Individual placeholder for missing child
                  placeholder_id = self._get_placeholder_id(child_id)
                  valid_children.append(placeholder_id)
                  placeholder_comp = {
                      "id": placeholder_id,
                      **self._placeholder_component,
                  }
                  # Avoid duplicates in extra_components
                  if not any(ec["id"] == placeholder_id for ec in extra_components):
                    extra_components.append(placeholder_comp)

              if not valid_children and field in ("children", "explicitList"):
                # If list is empty, check if it was partial in the buffer
                # (meaning it's a sequence that started but hasn't yielded items yet)
                term = f'"{field}"'
                if term in self._json_buffer:
                  # Simple check: is there a [ after the field name in the buffer?
                  after_field = self._json_buffer.split(term)[-1]
                  if "[" in after_field and "]" not in after_field.split("[")[0]:
                    placeholder_id = f"loading_children_{comp_id}"
                    valid_children.append(placeholder_id)
                    placeholder_comp = {
                        "id": placeholder_id,
                        **self._placeholder_component,
                    }
                    if not any(ec["id"] == placeholder_id for ec in extra_components):
                      extra_components.append(placeholder_comp)
              obj[field] = valid_children
            elif isinstance(obj[field], str):
              child_id = obj[field]
              if child_id not in self._seen_components:
                placeholder_id = self._get_placeholder_id(child_id)
                obj[field] = placeholder_id
                placeholder_comp = {
                    "id": placeholder_id,
                    **self._placeholder_component,
                }
                if not any(ec["id"] == placeholder_id for ec in extra_components):
                  extra_components.append(placeholder_comp)

        # Continue traversal on values
        for k, v in list(obj.items()):
          traverse(v, parent_key=k)
      elif isinstance(obj, list):
        for item in obj:
          traverse(item, parent_key)

    # Start recursion from the component content
    if isinstance(comp.get("component"), dict):
      traverse(comp.get("component", {}))
    else:
      # Flat style properties are siblings to 'component' type key
      traverse(comp)



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import copy
from unittest.mock import MagicMock
import pytest
from a2ui.schema.constants import (
    A2UI_OPEN_TAG,
    A2UI_CLOSE_TAG,
    VERSION_0_8,
    SURFACE_ID_KEY,
    CATALOG_COMPONENTS_KEY,
)
from a2ui.parser.constants import (
    MSG_TYPE_SURFACE_UPDATE,
    MSG_TYPE_BEGIN_RENDERING,
    MSG_TYPE_DELETE_SURFACE,
    MSG_TYPE_DATA_MODEL_UPDATE,
)
from a2ui.schema.catalog import A2uiCatalog
from a2ui.parser.streaming import A2uiStreamParser
from a2ui.parser.response_part import ResponsePart


@pytest.fixture
def mock_catalog():
  s2c_schema = {
      "title": "A2UI Message Schema",
      "type": "object",
      "additionalProperties": False,
      "properties": {
          "beginRendering": {
              "type": "object",
              "properties": {
                  "surfaceId": {"type": "string"},
                  "root": {"type": "string"},
              },
              "required": ["surfaceId", "root"],
          },
          "surfaceUpdate": {
              "type": "object",
              "properties": {
                  "surfaceId": {
                      "type": "string",
                  },
                  "components": {
                      "type": "array",
                      "items": {
                          "type": "object",
                          "properties": {
                              "id": {
                                  "type": "string",
                              },
                              "component": {
                                  "type": "object",
                                  "additionalProperties": True,
                              },
                          },
                          "required": ["id", "component"],
                      },
                  },
              },
              "required": ["surfaceId", "components"],
          },
          "dataModelUpdate": {
              "type": "object",
              "properties": {
                  "surfaceId": {"type": "string"},
                  "contents": {
                      "type": "array",
                      "items": {
                          "type": "object",
                          "properties": {
                              "key": {"type": "string"},
                              "valueString": {"type": "string"},
                              "valueNumber": {"type": "number"},
                              "valueBoolean": {"type": "boolean"},
                              "valueMap": {
                                  "type": "array",
                                  "items": {
                                      "type": "object",
                                      "properties": {
                                          "key": {"type": "string"},
                                          "valueString": {"type": "string"},
                                          "valueNumber": {"type": "number"},
                                          "valueBoolean": {"type": "boolean"},
                                      },
                                      "required": ["key"],
                                  },
                              },
                          },
                          "required": ["key"],
                      },
                  },
              },
              "required": ["surfaceId", "contents"],
          },
          "deleteSurface": {
              "type": "object",
              "properties": {
                  "surfaceId": {
                      "type": "string",
                  }
              },
              "required": ["surfaceId"],
          },
      },
  }
  catalog_schema = {
      "catalogId": "test_catalog",
      "components": {
          "Container": {
              "type": "object",
              "properties": {
                  "children": {
                      "type": "array",
                      "items": {"type": "string", "title": "ComponentId"},
                  }
              },
              "additionalProperties": True,
          },
          "Card": {
              "type": "object",
              "properties": {
                  "child": {"type": "string", "title": "ComponentId"},
                  "children": {
                      "type": "array",
                      "items": {"type": "string", "title": "ComponentId"},
                  },
              },
              "additionalProperties": True,
          },
          "Text": {"type": "object", "additionalProperties": True},
          "Loading": {"type": "object", "additionalProperties": True},
          "List": {"type": "object", "additionalProperties": True},
          "Row": {
              "type": "object",
              "properties": {
                  "children": {
                      "type": "object",
                      "properties": {
                          "explicitList": {
                              "type": "array",
                              "items": {"type": "string"},
                          },
                          "required": ["componentId", "dataBinding"],
                      },
                  }
              },
              "required": ["children"],
          },
          "Column": {
              "type": "object",
              "properties": {
                  "children": {
                      "type": "object",
                      "properties": {
                          "explicitList": {
                              "type": "array",
                              "items": {"type": "string", "title": "ComponentId"},
                          }
                      },
                  }
              },
          },
          "AudioPlayer": {
              "type": "object",
              "properties": {
                  "url": {
                      "type": "object",
                      "properties": {
                          "literalString": {"type": "string"},
                          "path": {"type": "string"},
                      },
                  },
              },
              "required": ["url"],
          },
      },
  }
  common_types_schema = {
      "$id": "https://a2ui.org/specification/v0_8/common_types.json",
      "type": "object",
      "$defs": {},
  }
  return A2uiCatalog(
      version=VERSION_0_8,
      name="test_catalog",
      s2c_schema=s2c_schema,
      common_types_schema=common_types_schema,
      catalog_schema=catalog_schema,
  )


def _normalize_messages(messages):
  """Sorts components in messages for stable comparison."""
  # Support ResponsePart list by extracting a2ui_json
  res = []
  for m in messages:
    if isinstance(m, ResponsePart):
      if m.a2ui_json:
        if isinstance(m.a2ui_json, list):
          res.extend(copy.deepcopy(m.a2ui_json))
        else:
          res.append(copy.deepcopy(m.a2ui_json))
    else:
      res.append(copy.deepcopy(m))

  for msg in res:
    if MSG_TYPE_SURFACE_UPDATE in msg:
      payload = msg[MSG_TYPE_SURFACE_UPDATE]
      if CATALOG_COMPONENTS_KEY in payload:
        payload[CATALOG_COMPONENTS_KEY].sort(key=lambda x: x.get("id", ""))
  return res


def assertResponseContainsMessages(response, expected_messages):
  """Asserts that the response parts contain the expected messages."""
  assert _normalize_messages(response) == _normalize_messages(expected_messages)


def assertResponseContainsNoA2UI(response):
  assert len(response) == 0 or response[0].a2ui_json == None


def assertResponseContainsText(response, expected_text):
  """Asserts that the response parts contain the expected text."""
  assert any(
      (p.text if isinstance(p, ResponsePart) else p) == expected_text for p in response
  )


def test_add_msg_type_deduplication():
  parser = A2uiStreamParser()
  parser.add_msg_type(MSG_TYPE_SURFACE_UPDATE)
  parser.add_msg_type(MSG_TYPE_SURFACE_UPDATE)
  assert parser.msg_types == [MSG_TYPE_SURFACE_UPDATE]

  parser.add_msg_type(MSG_TYPE_BEGIN_RENDERING)
  assert parser.msg_types == [MSG_TYPE_SURFACE_UPDATE, MSG_TYPE_BEGIN_RENDERING]
  parser.add_msg_type(MSG_TYPE_SURFACE_UPDATE)
  assert parser.msg_types == [MSG_TYPE_SURFACE_UPDATE, MSG_TYPE_BEGIN_RENDERING]


def test_streaming_msg_type_deduplication(mock_catalog):
  parser = A2uiStreamParser(catalog=mock_catalog)
  # 1. Send partial chunk that triggers sniffing
  chunk1 = A2UI_OPEN_TAG + '[{"surfaceUpdate": {"surfaceId": "s1", "components": ['
  parser.process_chunk(chunk1)

  # Sniffing should have added surfaceUpdate
  assert MSG_TYPE_SURFACE_UPDATE in parser.msg_types
  assert parser.msg_types.count(MSG_TYPE_SURFACE_UPDATE) == 1

  # 2. Send the rest, which triggers handle_complete_object
  chunk2 = (
      '{"id": "root", "component": {"Text": {"text": "hi"}}}]}]'
      f" {A2UI_CLOSE_TAG}"
  )
  parser.process_chunk(chunk2)

  # After completion, msg_types is reset
  assert parser.msg_types == []


def test_v08_path_heuristic_adds_slash(mock_catalog):
  """Tests that v0.8 adds a leading slash to relative paths."""
  parser = A2uiStreamParser(catalog=mock_catalog)
  # Disable validation for simplicity
  parser._validator = None

  # 1. Send beginRendering first to avoid buffering
  chunk_br = (
      A2UI_OPEN_TAG
      + '[{"beginRendering": {"surfaceId": "s1", "root": "root"}}]'
      + A2UI_CLOSE_TAG
  )
  list(parser.process_chunk(chunk_br))

  # 2. Send surfaceUpdate with a relative path
  chunk_su = (
      A2UI_OPEN_TAG
      + '[{"surfaceUpdate": {"surfaceId": "s1", "components": [{"id": "root",'
      ' "component": {"Text": {"text": {"path": "some/relative/path"}}}}]}}]'
      + A2UI_CLOSE_TAG
  )

  messages = []
  for part in parser.process_chunk(chunk_su):
    if part.a2ui_json:
      messages.extend(part.a2ui_json)

  # The path should have been prefixed with a slash
  assert len(messages) > 0
  comp = messages[0][MSG_TYPE_SURFACE_UPDATE]["components"][0]
  assert comp["component"]["Text"]["text"]["path"] == "/some/relative/path"



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..schema.constants import CATALOG_SCHEMA_KEY, VERSION_0_8, VERSION_0_9, VERSION_0_9_1

BASIC_CATALOG_NAME = "basic"

# Maps version to the relative path of the basic catalog schema in the source repo
BASIC_CATALOG_PATHS = {
    VERSION_0_8: {
        CATALOG_SCHEMA_KEY: "specification/v0_8/json/standard_catalog_definition.json"
    },
    VERSION_0_9: {CATALOG_SCHEMA_KEY: "specification/v0_9/catalogs/basic/catalog.json"},
    VERSION_0_9_1: {
        CATALOG_SCHEMA_KEY: "specification/v0_9_1/catalogs/basic/catalog.json"
    },
}



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import json
import logging
import os
import importlib.resources
from typing import Any, Optional, Callable
from dataclasses import dataclass, field
from .utils import load_from_bundled_resource
from ..inference_strategy import InferenceStrategy
from .constants import *
from .catalog import CatalogConfig, A2uiCatalog


class A2uiSchemaManager(InferenceStrategy):
  """Manages A2UI schema levels and prompt injection."""

  def __init__(
      self,
      version: str,
      catalogs: Optional[list[CatalogConfig]] = None,
      accepts_inline_catalogs: bool = False,
      schema_modifiers: Optional[
          list[Callable[[dict[str, Any]], dict[str, Any]]]
      ] = None,
  ):
    self._version = version
    self._accepts_inline_catalogs = accepts_inline_catalogs

    self._server_to_client_schema = None
    self._common_types_schema = None
    self._supported_catalogs: list[A2uiCatalog] = []
    self._catalog_example_paths: dict[str, str] = {}
    self._schema_modifiers = schema_modifiers or []
    self._load_schemas(version, catalogs or [])

  @property
  def accepts_inline_catalogs(self) -> bool:
    return self._accepts_inline_catalogs

  @property
  def supported_catalog_ids(self) -> list[str]:
    return [c.catalog_id for c in self._supported_catalogs]

  def _apply_modifiers(self, schema: dict[str, Any]) -> dict[str, Any]:
    if self._schema_modifiers:
      for modifier in self._schema_modifiers:
        schema = modifier(schema)
    return schema

  def _load_schemas(
      self,
      version: str,
      catalogs: Optional[list[CatalogConfig]] = None,
  ):
    """Loads separate schema components and processes catalogs."""
    catalogs = catalogs or []
    if version not in SPEC_VERSION_MAP:
      raise ValueError(
          f"Unknown A2UI specification version: {version}. Supported:"
          f" {list(SPEC_VERSION_MAP.keys())}"
      )

    # Load server-to-client and common types schemas
    self._server_to_client_schema = self._apply_modifiers(
        load_from_bundled_resource(
            version, SERVER_TO_CLIENT_SCHEMA_KEY, SPEC_VERSION_MAP
        )
    )
    self._common_types_schema = self._apply_modifiers(
        load_from_bundled_resource(version, COMMON_TYPES_SCHEMA_KEY, SPEC_VERSION_MAP)
    )

    # Process catalogs
    for config in catalogs:
      catalog_schema = config.provider.load()
      catalog_schema = self._apply_modifiers(catalog_schema)
      catalog = A2uiCatalog(
          version=version,
          name=config.name,
          catalog_schema=catalog_schema,
          s2c_schema=self._server_to_client_schema,
          common_types_schema=self._common_types_schema,
      )
      self._supported_catalogs.append(catalog)
      self._catalog_example_paths[catalog.catalog_id] = config.examples_path

  def _select_catalog(
      self, client_ui_capabilities: Optional[dict[str, Any]] = None
  ) -> A2uiCatalog:
    """Selects the component catalog for the prompt based on client capabilities.

    Selection priority:
    1. If inline catalogs are provided (and accepted by the agent), their
       components are merged on top of a base catalog. The base is determined
       by supportedCatalogIds (if also provided) or the agent's default catalog.
    2. If only supportedCatalogIds is provided, pick the first mutually
       supported catalog.
    3. Fallback to the first agent-supported catalog (usually the bundled catalog).

    Args:
      client_ui_capabilities: A dictionary of client UI capabilities, containing
        inline catalogs and client-supported catalog IDs.

    Returns:
      The resolved A2uiCatalog.
    Raises:
      ValueError: If inline catalogs are sent but not accepted, or if no
        mutually supported catalog is found.
    """
    if not self._supported_catalogs:
      raise ValueError("No supported catalogs found.")  # This should not happen.

    if not client_ui_capabilities or not isinstance(client_ui_capabilities, dict):
      return self._supported_catalogs[0]

    inline_catalogs: list[dict[str, Any]] = client_ui_capabilities.get(
        INLINE_CATALOGS_KEY, []
    )
    client_supported_catalog_ids: list[str] = client_ui_capabilities.get(
        SUPPORTED_CATALOG_IDS_KEY, []
    )

    if not self._accepts_inline_catalogs and inline_catalogs:
      raise ValueError(
          f"Inline catalog '{INLINE_CATALOGS_KEY}' is provided in client UI"
          " capabilities. However, the agent does not accept inline catalogs."
      )

    if inline_catalogs:
      # Determine the base catalog: use supportedCatalogIds if provided,
      # otherwise fall back to the agent's default catalog.
      base_catalog = self._supported_catalogs[0]
      if client_supported_catalog_ids:
        agent_supported_catalogs = {c.catalog_id: c for c in self._supported_catalogs}
        for cscid in client_supported_catalog_ids:
          if cscid in agent_supported_catalogs:
            base_catalog = agent_supported_catalogs[cscid]
            break

      merged_schema = copy.deepcopy(base_catalog.catalog_schema)

      for inline_catalog_schema in inline_catalogs:
        inline_catalog_schema = self._apply_modifiers(inline_catalog_schema)
        inline_components = inline_catalog_schema.get(CATALOG_COMPONENTS_KEY, {})
        merged_schema[CATALOG_COMPONENTS_KEY].update(inline_components)

      return A2uiCatalog(
          version=self._version,
          name=INLINE_CATALOG_NAME,
          catalog_schema=merged_schema,
          s2c_schema=self._server_to_client_schema,
          common_types_schema=self._common_types_schema,
      )

    if not client_supported_catalog_ids:
      return self._supported_catalogs[0]

    agent_supported_catalogs = {c.catalog_id: c for c in self._supported_catalogs}
    for cscid in client_supported_catalog_ids:
      if cscid in agent_supported_catalogs:
        return agent_supported_catalogs[cscid]

    raise ValueError(
        "No client-supported catalog found on the agent side. Agent-supported catalogs"
        f" are: {[c.catalog_id for c in self._supported_catalogs]}"
    )

  def get_selected_catalog(
      self,
      client_ui_capabilities: Optional[dict[str, Any]] = None,
      allowed_components: Optional[list[str]] = None,
      allowed_messages: Optional[list[str]] = None,
  ) -> A2uiCatalog:
    """Gets the selected catalog after selection and component pruning."""
    catalog = self._select_catalog(client_ui_capabilities)
    pruned_catalog = catalog.with_pruning(allowed_components, allowed_messages)
    return pruned_catalog

  def load_examples(self, catalog: A2uiCatalog, validate: bool = False) -> str:
    """Loads examples for a catalog."""
    if catalog.catalog_id in self._catalog_example_paths:
      return catalog.load_examples(
          self._catalog_example_paths[catalog.catalog_id], validate=validate
      )
    return ""

  def generate_system_prompt(
      self,
      role_description: str,
      workflow_description: str = "",
      ui_description: str = "",
      client_ui_capabilities: Optional[dict[str, Any]] = None,
      allowed_components: Optional[list[str]] = None,
      allowed_messages: Optional[list[str]] = None,
      include_schema: bool = False,
      include_examples: bool = False,
      validate_examples: bool = False,
  ) -> str:
    """Assembles the final system instruction for the LLM."""
    parts = [role_description]

    workflow = DEFAULT_WORKFLOW_RULES
    if workflow_description:
      workflow += f"\n{workflow_description}"
    parts.append(f"## Workflow Description:\n{workflow}")

    if ui_description:
      parts.append(f"## UI Description:\n{ui_description}")

    selected_catalog = self.get_selected_catalog(
        client_ui_capabilities, allowed_components, allowed_messages
    )

    if include_schema:
      parts.append(selected_catalog.render_as_llm_instructions())

    if include_examples:
      examples_str = self.load_examples(selected_catalog, validate=validate_examples)
      if examples_str:
        parts.append(f"### Examples:\n{examples_str}")

    return "\n\n".join(parts)



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import copy
import pytest
from unittest.mock import MagicMock
from a2ui.schema.manager import A2uiSchemaManager, A2uiCatalog, CatalogConfig
from a2ui.schema.common_modifiers import remove_strict_validation
from a2ui.schema.constants import VERSION_0_8, VERSION_0_9
from a2ui.schema.validator import (
    _find_root_id as find_root_id,
    extract_component_ref_fields,
    analyze_topology,
    get_component_references,
)


class TestValidator:

  @pytest.fixture
  def catalog_0_9(self):
    s2c_schema = {
        "$schema": "https://json-schema.org/draft/2020-12/schema",
        "$id": "https://a2ui.org/specification/v0_9/server_to_client.json",
        "title": "A2UI Message Schema",
        "oneOf": [
            {"$ref": "#/$defs/CreateSurfaceMessage"},
            {"$ref": "#/$defs/UpdateComponentsMessage"},
            {"$ref": "#/$defs/UpdateDataModelMessage"},
            {"$ref": "#/$defs/DeleteSurfaceMessage"},
        ],
        "$defs": {
            "CreateSurfaceMessage": {
                "type": "object",
                "properties": {
                    "version": {"const": "v0.9"},
                    "createSurface": {
                        "type": "object",
                        "properties": {
                            "surfaceId": {
                                "type": "string",
                            },
                            "catalogId": {
                                "type": "string",
                            },
                            "theme": {
                                "type": "object",
                                "additionalProperties": True,
                            },
                        },
                        "required": ["surfaceId", "catalogId"],
                        "additionalProperties": False,
                    },
                },
                "required": ["version", "createSurface"],
                "additionalProperties": False,
            },
            "UpdateComponentsMessage": {
                "type": "object",
                "properties": {
                    "version": {"const": "v0.9"},
                    "updateComponents": {
                        "type": "object",
                        "properties": {
                            "surfaceId": {
                                "type": "string",
                            },
                            "components": {
                                "type": "array",
                                "minItems": 1,
                                "items": {"$ref": "catalog.json#/$defs/anyComponent"},
                            },
                        },
                        "required": ["surfaceId", "components"],
                        "additionalProperties": False,
                    },
                },
                "required": ["version", "updateComponents"],
                "additionalProperties": False,
            },
            "UpdateDataModelMessage": {
                "type": "object",
                "properties": {
                    "version": {"const": "v0.9"},
                    "updateDataModel": {
                        "type": "object",
                        "properties": {
                            "surfaceId": {
                                "type": "string",
                            },
                            "path": {"type": "string"},
                            "value": {"additionalProperties": True},
                        },
                        "required": ["surfaceId"],
                        "additionalProperties": False,
                    },
                },
                "required": ["version", "updateDataModel"],
                "additionalProperties": False,
            },
            "DeleteSurfaceMessage": {
                "type": "object",
                "properties": {
                    "version": {"const": "v0.9"},
                    "deleteSurface": {
                        "type": "object",
                        "properties": {
                            "surfaceId": {"type": "string"},
                        },
                        "required": ["surfaceId"],
                        "additionalProperties": False,
                    },
                },
                "required": ["deleteSurface", "version"],
                "additionalProperties": False,
            },
        },
    }
    catalog_schema = {
        "$schema": "https://json-schema.org/draft/2020-12/schema",
        "$id": "https://a2ui.org/specification/v0_9/catalogs/basic/catalog.json",
        "title": "A2UI Basic Catalog",
        "catalogId": "https://a2ui.dev/specification/v0_9/catalogs/basic/catalog.json",
        "components": {
            "Text": {
                "type": "object",
                "allOf": [
                    {"$ref": "common_types.json#/$defs/ComponentCommon"},
                    {"$ref": "#/$defs/CatalogComponentCommon"},
                ],
                "properties": {
                    "component": {"const": "Text"},
                    "text": {"$ref": "common_types.json#/$defs/DynamicString"},
                },
                "required": ["component", "text"],
                "unevaluatedProperties": False,
            },
            "Image": {
                "type": "object",
                "allOf": [
                    {"$ref": "common_types.json#/$defs/ComponentCommon"},
                    {"$ref": "#/$defs/CatalogComponentCommon"},
                ],
                "properties": {
                    "component": {"const": "Image"},
                    "url": {"type": "string"},
                },
                "required": ["component", "url"],
                "unevaluatedProperties": False,
            },
            "Icon": {
                "type": "object",
                "allOf": [
                    {"$ref": "common_types.json#/$defs/ComponentCommon"},
                    {"$ref": "#/$defs/CatalogComponentCommon"},
                ],
                "properties": {
                    "component": {"const": "Icon"},
                    "name": {"type": "string"},
                },
                "required": ["component", "name"],
                "unevaluatedProperties": False,
            },
            "Column": {
                "type": "object",
                "allOf": [
                    {"$ref": "common_types.json#/$defs/ComponentCommon"},
                    {"$ref": "#/$defs/CatalogComponentCommon"},
                ],
                "properties": {
                    "component": {"const": "Column"},
                    "children": {"$ref": "common_types.json#/$defs/ChildList"},
                },
                "required": ["component", "children"],
                "unevaluatedProperties": False,
            },
            "Card": {
                "type": "object",
                "allOf": [
                    {"$ref": "common_types.json#/$defs/ComponentCommon"},
                    {"$ref": "#/$defs/CatalogComponentCommon"},
                ],
                "properties": {
                    "component": {"const": "Card"},
                    "child": {"$ref": "common_types.json#/$defs/ComponentId"},
                },
                "required": ["component", "child"],
                "unevaluatedProperties": False,
            },
            "Button": {
                "type": "object",
                "allOf": [
                    {"$ref": "common_types.json#/$defs/ComponentCommon"},
                    {"$ref": "#/$defs/CatalogComponentCommon"},
                ],
                "properties": {
                    "component": {"const": "Button"},
                    "text": {"type": "string"},
                    "action": {"$ref": "common_types.json#/$defs/Action"},
                },
                "required": ["component", "text", "action"],
                "unevaluatedProperties": False,
            },
            "List": {
                "type": "object",
                "allOf": [
                    {"$ref": "common_types.json#/$defs/ComponentCommon"},
                    {"$ref": "#/$defs/CatalogComponentCommon"},
                    {
                        "type": "object",
                        "properties": {
                            "component": {"const": "List"},
                            "children": {"$ref": "common_types.json#/$defs/ChildList"},
                            "direction": {
                                "type": "string",
                                "enum": ["vertical", "horizontal"],
                            },
                        },
                        "required": ["component", "children"],
                    },
                ],
                "unevaluatedProperties": False,
            },
        },
        "$defs": {
            "CatalogComponentCommon": {
                "type": "object",
                "properties": {"weight": {"type": "number"}},
            },
            "anyComponent": {
                "oneOf": [
                    {"$ref": "#/components/Text"},
                    {"$ref": "#/components/Image"},
                    {"$ref": "#/components/Icon"},
                    {"$ref": "#/components/Column"},
                    {"$ref": "#/components/Card"},
                    {"$ref": "#/components/Button"},
                    {"$ref": "#/components/List"},
                ],
                "discriminator": {"propertyName": "component"},
            },
        },
    }
    common_types_schema = {
        "$schema": "https://json-schema.org/draft/2020-12/schema",
        "$id": "https://a2ui.org/specification/v0_9/common_types.json",
        "title": "A2UI Common Types",
        "$defs": {
            "ComponentId": {
                "type": "string",
            },
            "AccessibilityAttributes": {
                "type": "object",
                "properties": {
                    "label": {
                        "$ref": "#/$defs/DynamicString",
                    }
                },
            },
            "Action": {"type": "object", "additionalProperties": True},
            "ComponentCommon": {
                "type": "object",
                "properties": {"id": {"$ref": "#/$defs/ComponentId"}},
                "required": ["id"],
            },
            "DataBinding": {"type": "object"},
            "DynamicString": {
                "anyOf": [{"type": "string"}, {"$ref": "#/$defs/DataBinding"}]
            },
            "DynamicValue": {
                "anyOf": [
                    {"type": "object"},
                    {"type": "array"},
                    {"$ref": "#/$defs/DataBinding"},
                ]
            },
            "DynamicNumber": {
                "anyOf": [{"type": "number"}, {"$ref": "#/$defs/DataBinding"}]
            },
            "ChildList": {
                "oneOf": [
                    {"type": "array", "items": {"$ref": "#/$defs/ComponentId"}},
                    {
                        "type": "object",
                        "properties": {
                            "componentId": {"$ref": "#/$defs/ComponentId"},
                            "path": {"type": "string"},
                        },
                        "required": ["componentId", "path"],
                        "additionalProperties": False,
                    },
                ]
            },
        },
    }
    return A2uiCatalog(
        version=VERSION_0_9,
        name="standard",
        catalog_schema=catalog_schema,
        s2c_schema=s2c_schema,
        common_types_schema=common_types_schema,
    )

  @pytest.fixture
  def catalog_0_8(self):
    s2c_schema = {
        "title": "A2UI Message Schema",
        "description": "Describes a JSON payload for an A2UI message.",
        "type": "object",
        "additionalProperties": False,
        "properties": {
            "beginRendering": {
                "type": "object",
                "additionalProperties": False,
                "properties": {
                    "surfaceId": {"type": "string"},
                    "root": {"type": "string"},
                    "styles": {
                        "type": "object",
                        "description": "Styling information for the UI.",
                        "additionalProperties": True,
                    },
                },
                "required": ["surfaceId"],
            },
            "surfaceUpdate": {
                "type": "object",
                "additionalProperties": False,
                "properties": {
                    "surfaceId": {
                        "type": "string",
                    },
                    "components": {
                        "type": "array",
                        "items": {
                            "type": "object",
                            "additionalProperties": False,
                            "properties": {
                                "id": {
                                    "type": "string",
                                },
                                "component": {
                                    "type": "object",
                                    "description": "A wrapper object.",
                                    "additionalProperties": True,
                                },
                            },
                            "required": ["id", "component"],
                        },
                    },
                },
                "required": ["surfaceId", "components"],
            },
            "dataModelUpdate": {
                "type": "object",
                "properties": {
                    "surfaceId": {"type": "string"},
                    "contents": {"type": "object", "additionalProperties": True},
                },
            },
        },
        "additionalProperties": False,
    }
    catalog_schema = {
        "catalogId": (
            "https://a2ui.org/specification/v0_8/json/standard_catalog_definition.json"
        ),
        "components": {
            "Column": {
                "type": "object",
                "additionalProperties": True,
                "properties": {
                    "children": {"type": "array", "items": {"type": "string"}}
                },
            },
            "Card": {
                "type": "object",
                "additionalProperties": True,
                "properties": {"child": {"type": "string"}},
            },
            "Button": {
                "type": "object",
                "additionalProperties": True,
                "properties": {
                    "label": {"type": "string"},
                    "action": {
                        "type": "object",
                        "properties": {
                            "functionCall": {
                                "type": "object",
                                "properties": {
                                    "call": {"type": "string"},
                                    "args": {"type": "object"},
                                },
                            }
                        },
                    },
                },
            },
            "Text": {
                "type": "object",
                "additionalProperties": True,
                "properties": {
                    "text": {
                        "anyOf": [
                            {"type": "string"},
                            {"type": "object", "additionalProperties": True},
                        ]
                    }
                },
            },
            "List": {
                "type": "object",
                "additionalProperties": True,
            },
        },
        "styles": {"font": {"type": "string"}, "primaryColor": {"type": "string"}},
    }
    return A2uiCatalog(
        version=VERSION_0_8,
        name="standard",
        catalog_schema=catalog_schema,
        s2c_schema=s2c_schema,
        common_types_schema=None,
    )

  @pytest.fixture(params=[VERSION_0_8, VERSION_0_9])
  def test_catalog(self, request, catalog_0_8, catalog_0_9):
    """Parameterized fixture to run tests on both v0.8 and v0.9 catalogs."""
    if request.param == VERSION_0_8:
      return catalog_0_8
    return catalog_0_9

  def test_pretty_error_messages(self, catalog_0_9):
    payload = [
        {
            "version": "v0.9",
            "createSurface": {
                "surfaceId": "recipe-card",
                "catalogId": (
                    "https://a2ui.org/specification/v0_9/catalogs/basic/catalog.json"
                ),
            },
        },
        {
            "version": "v0.9",
            "updateComponents": {
                "surfaceId": "recipe-card",
                "components": [
                    {
                        "id": "main-column",
                        "component": "Column",
                        "children": ["recipe-image"],
                        "gap": "small",
                    },
                    {
                        "id": "recipe-image",
                        "component": "Image",
                        "url": {"path": "/image"},
                        "altText": {"path": "/title"},
                        "fit": "cover",
                    },
                    {
                        "id": "title",
                        "component": "Text",
                        "text": {"path": "/title"},
                        "usageHint": "h3",
                    },
                    {
                        "id": "rating-row",
                        "component": "Row",
                        "children": ["star-icon"],
                    },
                ],
            },
        },
        {
            "version": "v0.9",
            "updateDataModel": {
                "surfaceId": "recipe-card",
                "value": {
                    "image": (
                        "https://images.unsplash.com/photo-1546069901-ba9599a7e63c?w=300&h=180&fit=crop"
                    )
                },
            },
        },
        {"version": "v0.9", "deleteSurface": {}},
        {"unknownMessage": {}},
    ]

    with pytest.raises(ValueError) as excinfo:
      catalog_0_9.validator.validate(payload)

    err_text = str(excinfo.value)
    print(f"\nVALIDATOR_OUTPUT_START\n{err_text}\nVALIDATOR_OUTPUT_END")

    assert "Unknown component: Row" in err_text
    assert "'usageHint' was unexpected" in err_text
    assert "'gap' was unexpected" in err_text
    assert "'altText', 'fit' were unexpected" in err_text
    assert "'surfaceId' is a required property" in err_text
    assert "{'path': '/image'} is not of type 'string'" in err_text
    assert "Unknown message type with keys ['unknownMessage']" in err_text

  def test_bundle_0_8(self, catalog_0_8):
    bundled = catalog_0_8.validator._bundle_0_8_schemas()

    # Verify styles injection
    styles_node = bundled["properties"]["beginRendering"]["properties"]["styles"]
    assert styles_node["additionalProperties"] is False
    assert "font" in styles_node["properties"]
    assert "primaryColor" in styles_node["properties"]

    # Verify component injection
    component_node = bundled["properties"]["surfaceUpdate"]["properties"]["components"][
        "items"
    ]["properties"]["component"]
    assert component_node["additionalProperties"] is False
    assert "Text" in component_node["properties"]
    assert "Button" in component_node["properties"]

  def test_find_root_id_v08(self):
    messages = [
        {"beginRendering": {"surfaceId": "s1", "root": "custom-root"}},
        {"surfaceUpdate": {"surfaceId": "s1", "components": []}},
    ]
    assert find_root_id(messages) == "custom-root"

  def test_find_root_id_v09(self):
    # For v0.9, if createSurface is provided, the root is 'root'
    messages = [
        {"createSurface": {"surfaceId": "s1"}},
        {"updateComponents": {"surfaceId": "s1", "components": []}},
    ]
    assert find_root_id(messages) == "root"

    # For an incremental update, there is no root
    messages = [{"updateComponents": {"surfaceId": "s1", "components": []}}]
    assert find_root_id(messages) is None

  def test_get_component_references(self):
    # Mock ref_fields_map
    ref_fields_map = {
        "Container": ({"child"}, {"children"}),
        "Text": (set(), set()),
    }
    comp = {
        "id": "c1",
        "component": "Container",
        "child": "c2",
        "children": ["c3", "c4"],
    }
    refs = list(get_component_references(comp, ref_fields_map))
    assert ("c2", "child") in refs
    assert ("c3", "children") in refs
    assert ("c4", "children") in refs

  def test_analyze_topology_circular(self):
    ref_fields_map = {"Node": ({"next"}, set())}
    components = [
        {"id": "c1", "component": "Node", "next": "c2"},
        {"id": "c2", "component": "Node", "next": "c1"},
    ]
    with pytest.raises(ValueError, match="Circular reference detected"):
      analyze_topology("c1", components, ref_fields_map)

  def test_analyze_topology_self_ref(self):
    ref_fields_map = {"Node": ({"next"}, set())}
    components = [{"id": "c1", "component": "Node", "next": "c1"}]
    with pytest.raises(ValueError, match="Self-reference detected"):
      analyze_topology("c1", components, ref_fields_map)

  def test_analyze_topology_reachable(self):
    ref_fields_map = {"Node": ({"next"}, set())}
    components = [
        {"id": "root", "component": "Node", "next": "c1"},
        {"id": "c1", "component": "Node", "next": "c2"},
        {"id": "c2", "component": "Node"},
        {"id": "orphan", "component": "Node"},
    ]
    reachable = analyze_topology("root", components, ref_fields_map)
    assert reachable == {"root", "c1", "c2"}

  def test_extract_component_ref_fields_mock(self):
    # Test with a mock catalog
    catalog = MagicMock(spec=A2uiCatalog)
    catalog.version = VERSION_0_9
    catalog.common_types_schema = {
        "$defs": {
            "ComponentId": {"type": "string"},
            "ChildList": {
                "oneOf": [
                    {"type": "array", "items": {"$ref": "#/$defs/ComponentId"}},
                    {
                        "type": "object",
                        "properties": {"componentId": {"$ref": "#/$defs/ComponentId"}},
                    },
                ]
            },
        }
    }
    catalog.catalog_schema = {
        "components": {
            "MyComp": {
                "properties": {
                    "ref": {"$ref": "common_types.json#/$defs/ComponentId"},
                    "multi": {"$ref": "common_types.json#/$defs/ChildList"},
                }
            }
        }
    }
    # Mock s2c_schema to avoid errors in extraction
    catalog.s2c_schema = {}

    ref_map = extract_component_ref_fields(catalog)
    assert "MyComp" in ref_map
    single_refs, list_refs = ref_map["MyComp"]
    assert "ref" in single_refs
    assert "multi" in list_refs



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import json
from typing import Any, List, Dict, Optional, Set

from .streaming import A2uiStreamParser
from .response_part import ResponsePart
from .constants import *
from ..schema.constants import VERSION_0_9, SURFACE_ID_KEY, CATALOG_COMPONENTS_KEY


class A2uiStreamParserV09(A2uiStreamParser):
  """Streaming parser implementation for A2UI v0.9 specification."""

  def __init__(self, catalog=None):
    super().__init__(catalog=catalog)
    # v0.9 default root is "root"
    self._default_root_id = DEFAULT_ROOT_ID

  @property
  def _placeholder_component(self) -> Dict[str, Any]:
    """Returns a v0.9 flat style placeholder component specification."""
    return {
        'component': 'Row',
        'children': [],
    }

  @property
  def _data_model_msg_type(self) -> str:
    """Returns the message type identifier for data model updates."""
    return MSG_TYPE_UPDATE_DATA_MODEL

  def is_protocol_msg(self, obj: Dict[str, Any]) -> bool:
    """Checks if the object is a recognized v0.9 message."""
    return any(
        k in obj
        for k in (
            MSG_TYPE_CREATE_SURFACE,
            MSG_TYPE_UPDATE_COMPONENTS,
            MSG_TYPE_UPDATE_DATA_MODEL,
        )
    )

  def _sniff_metadata(self):
    """Sniffs for v0.9 metadata in the json_buffer."""

    def get_latest_value(key: str) -> Optional[str]:
      idx = len(self._json_buffer)
      while True:
        idx = self._json_buffer.rfind(f'"{key}"', 0, idx)
        if idx == -1:
          return None
        match = re.match(rf'"{key}"\s*:\s*"([^"]+)"', self._json_buffer[idx:])
        if match:
          return match.group(1)

    self.surface_id = get_latest_value('surfaceId')

    parsed_root = get_latest_value('root')
    if parsed_root is not None:
      self.root_id = parsed_root

    if f'"{MSG_TYPE_CREATE_SURFACE}":' in self._json_buffer:
      self.add_msg_type(MSG_TYPE_CREATE_SURFACE)
    if f'"{MSG_TYPE_UPDATE_COMPONENTS}":' in self._json_buffer:
      self.add_msg_type(MSG_TYPE_UPDATE_COMPONENTS)
    if f'"{MSG_TYPE_UPDATE_DATA_MODEL}":' in self._json_buffer:
      self.add_msg_type(MSG_TYPE_UPDATE_DATA_MODEL)

  def _handle_complete_object(
      self,
      obj: Dict[str, Any],
      sid: Optional[str],
      messages: List[ResponsePart],
  ) -> bool:
    """Handles v0.9 specific complete objects."""
    if not isinstance(obj, dict):
      return False

    if self._validator:
      self._validator.validate(obj, root_id=sid, strict_integrity=False)

    # Update state based on the message content
    surface_id = obj.get(SURFACE_ID_KEY, self.surface_id)
    if MSG_TYPE_UPDATE_COMPONENTS in obj:
      val = obj[MSG_TYPE_UPDATE_COMPONENTS]
      if isinstance(val, dict):
        surface_id = val.get(SURFACE_ID_KEY) or surface_id
    elif MSG_TYPE_CREATE_SURFACE in obj:
      val = obj[MSG_TYPE_CREATE_SURFACE]
      if isinstance(val, dict):
        surface_id = val.get(SURFACE_ID_KEY) or surface_id

    self.surface_id = surface_id
    sid = self.surface_id or 'unknown'

    # v0.9 Specific Handling
    if MSG_TYPE_CREATE_SURFACE in obj:
      val = obj[MSG_TYPE_CREATE_SURFACE]
      if isinstance(val, dict):
        self.root_id = val.get('root', self.root_id or DEFAULT_ROOT_ID)
      self._buffered_start_message = obj

      # Yield createSurface immediately when it completes
      if sid not in self._yielded_start_messages:
        self._yield_messages([obj], messages)
        self._yielded_start_messages.add(sid)
        self._yielded_surfaces_set.add(sid)
        self._buffered_start_message = None

      if sid in self._pending_messages:
        # Clear pending messages when createSurface arrives, we want a fresh start!
        self._pending_messages.pop(sid)

      self.yield_reachable(messages)
      return True

    if MSG_TYPE_UPDATE_COMPONENTS in obj:
      self.add_msg_type(MSG_TYPE_UPDATE_COMPONENTS)
      self.root_id = obj[MSG_TYPE_UPDATE_COMPONENTS].get(
          'root', self.root_id or DEFAULT_ROOT_ID
      )
      components = obj[MSG_TYPE_UPDATE_COMPONENTS].get('components', [])
      for comp in components:
        if isinstance(comp, dict) and 'id' in comp:
          self._seen_components[comp['id']] = comp
      self.yield_reachable(messages, check_root=True, raise_on_orphans=False)
      return True

    if MSG_TYPE_DELETE_SURFACE in obj:
      if sid not in self._yielded_start_messages:
        self._pending_messages.setdefault(sid, []).append(obj)
        return True
      self.add_msg_type(MSG_TYPE_DELETE_SURFACE)
      self._yield_messages([obj], messages)
      return True

    if MSG_TYPE_UPDATE_DATA_MODEL in obj:

      self.add_msg_type(MSG_TYPE_UPDATE_DATA_MODEL)
      self.update_data_model(obj[MSG_TYPE_UPDATE_DATA_MODEL], messages)
      self._yield_messages([obj], messages)
      return True

    return False

  def _construct_sniffed_data_model_message(
      self, active_msg_type: str, delta_msg_payload: Dict[str, Any]
  ) -> Dict[str, Any]:
    """Returns the message to yield for a partial data model update for v0.9."""
    return {'version': 'v0.9', active_msg_type: delta_msg_payload}

  def _sniff_partial_data_model(self, messages: List[ResponsePart]) -> None:
    """Sniffs for partial data model updates in v0.9 (value property)."""
    msg_type = MSG_TYPE_UPDATE_DATA_MODEL
    if f'"{msg_type}"' not in self._json_buffer:
      return

    for b_type, start_idx in reversed(self._brace_stack):
      if b_type != '{':
        continue
      raw_fragment = self._json_buffer[start_idx:]
      if not raw_fragment:
        continue

      fixed_fragment = self._fix_json(raw_fragment)
      obj = None
      try:
        obj = json.loads(fixed_fragment)
      except json.JSONDecodeError:
        # Fallback: iteratively strip from the last comma
        trimmed = raw_fragment
        while ',' in trimmed:
          trimmed = trimmed.rsplit(',', 1)[0]
          try:
            fixed_trimmed = self._fix_json(trimmed)
            if fixed_trimmed:
              obj = json.loads(fixed_trimmed)
              break
          except json.JSONDecodeError:
            continue

      if obj and isinstance(obj, dict) and msg_type in obj:

        dm_obj = obj[msg_type]
        if isinstance(dm_obj, dict) and 'value' in dm_obj:
          value_map = dm_obj['value']
          if isinstance(value_map, dict):
            # Find delta against yielded data model
            delta = {}
            for k, v in value_map.items():
              if self._yielded_data_model.get(k) != v:
                delta[k] = v

            if delta:
              sid = dm_obj.get(SURFACE_ID_KEY) or self._surface_id or 'default'
              delta_msg_payload = {
                  SURFACE_ID_KEY: sid,
                  'value': delta,
              }
              delta_msg = self._construct_sniffed_data_model_message(
                  msg_type, delta_msg_payload
              )
              self._yield_messages([delta_msg], messages, strict_integrity=False)
              # Do NOT update _yielded_data_model here, let update_data_model do it when complete
              # Wait! If we don't update it, will we over-yield it in the next chunk?
              # Yes, we might. So we should update it or track it!
              # The base class updates it (line 644 approx). So we should update it too!
              self._yielded_data_model.update(delta)

  def _construct_partial_message(
      self, processed_components: List[Dict[str, Any]], active_msg_type: str
  ) -> Dict[str, Any]:
    """Constructs a partial message for v0.9 (updateComponents)."""
    payload = {
        CATALOG_COMPONENTS_KEY: processed_components,
    }
    if self.surface_id:
      payload[SURFACE_ID_KEY] = self.surface_id
    return {'version': 'v0.9', MSG_TYPE_UPDATE_COMPONENTS: payload}

  @property
  def _yielded_surfaces_set(self) -> Set[str]:
    """Provides access to version-specific yielded surfaces set."""
    if not hasattr(self, '_yielded_create_surfaces'):
      self._yielded_create_surfaces: Set[str] = set()
    return self._yielded_create_surfaces

  def _get_active_msg_type_for_components(self) -> Optional[str]:
    """Determines which msg_type to use when wrapping component updates."""
    if self._active_msg_type:
      return self._active_msg_type
    for mt in self._msg_types:
      if mt in (MSG_TYPE_UPDATE_COMPONENTS, MSG_TYPE_CREATE_SURFACE):
        self._active_msg_type = mt
        return mt
    return self._msg_types[0] if self._msg_types else None

  def _deduplicate_data_model(self, m: Dict[str, Any], strict_integrity: bool) -> bool:
    if MSG_TYPE_UPDATE_DATA_MODEL in m:
      udm = m[MSG_TYPE_UPDATE_DATA_MODEL]
      if isinstance(udm, dict):
        is_new = False
        for k, v in udm.items():
          if k not in (SURFACE_ID_KEY, 'root') and self._yielded_data_model.get(k) != v:
            is_new = True
            break
        if not is_new and strict_integrity:
          return False
        # Update yielded model
        for k, v in udm.items():
          if k not in (SURFACE_ID_KEY, 'root'):
            self._yielded_data_model[k] = v
    return True



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib.util
import os
import sys
import shutil
from hatchling.builders.hooks.plugin.interface import BuildHookInterface


def load_module(project_root, rel_path, filename, module_name):
  """Loads a module directly from its path in src/."""
  path = os.path.join(project_root, "src", rel_path.replace(".", os.sep), filename)
  if not os.path.exists(path):
    raise RuntimeError(f"Could not find module at {path}")

  # Add src to sys.path so absolute imports work
  src_path = os.path.abspath(os.path.join(project_root, "src"))
  if src_path not in sys.path:
    sys.path.insert(0, src_path)

  spec = importlib.util.spec_from_file_location(module_name, path)
  if spec and spec.loader:
    module = importlib.util.module_from_spec(spec)
    # Set the package context to allow relative imports if any
    module.__package__ = rel_path
    sys.modules[module_name] = module
    spec.loader.exec_module(module)
    return module
  raise RuntimeError(f"Could not load module from {path}")


class PackSpecsBuildHook(BuildHookInterface):

  def initialize(self, version, build_data):
    project_root = self.root

    # Load constants and utils dynamically from src/
    schema_path = "a2ui.schema"
    a2ui_constants = load_module(
        project_root, schema_path, "constants.py", "_constants_load"
    )
    a2ui_utils = load_module(project_root, schema_path, "utils.py", "_utils_load")

    basic_catalog_constants = load_module(
        project_root,
        "a2ui.basic_catalog",
        "constants.py",
        "_basic_catalog_constants_load",
    )

    spec_version_map = a2ui_constants.SPEC_VERSION_MAP
    a2ui_asset_package = a2ui_constants.A2UI_ASSET_PACKAGE
    specification_dir = a2ui_constants.SPECIFICATION_DIR

    # Dynamically find repo root by looking for specification_dir
    repo_root = a2ui_utils.find_repo_root(project_root)
    if not repo_root:
      # Check for PKG-INFO which implies a packaged state (sdist).
      # If PKG-INFO is present, trust the bundled assets.
      if os.path.exists(os.path.join(project_root, "PKG-INFO")):
        print("Repository root not found, but PKG-INFO present (sdist). Skipping copy.")
        return

      raise RuntimeError(
          f"Could not find repository root (looked for '{specification_dir}'"
          " directory)."
      )

    # Target directory: src/a2ui/assets
    target_base = os.path.join(
        project_root, "src", a2ui_asset_package.replace(".", os.sep)
    )

    self._pack_schemas(repo_root, spec_version_map, target_base)
    self._pack_basic_catalogs(
        repo_root, basic_catalog_constants.BASIC_CATALOG_PATHS, target_base
    )

  def _pack_schemas(self, repo_root, spec_map, target_base):
    for ver, schema_map in spec_map.items():
      target_dir = os.path.join(target_base, ver)
      os.makedirs(target_dir, exist_ok=True)

      for _schema_key, source_rel_path in schema_map.items():
        self._copy_schema(repo_root, source_rel_path, target_dir)

  def _pack_basic_catalogs(self, repo_root, catalog_paths, target_base):
    for ver, path_map in catalog_paths.items():
      target_dir = os.path.join(target_base, ver)
      os.makedirs(target_dir, exist_ok=True)

      for _key, source_rel_path in path_map.items():
        self._copy_schema(repo_root, source_rel_path, target_dir)

  def _copy_schema(self, repo_root, source_rel_path, target_dir):
    source_path = os.path.join(repo_root, source_rel_path)

    if not os.path.exists(source_path):
      print(
          f"WARNING: Source schema file not found at {source_path}. Build"
          " might produce incomplete wheel if not running from monorepo"
          " root."
      )
      return

    filename = os.path.basename(source_path)
    dst_file = os.path.join(target_dir, filename)

    print(f"Copying {source_path} -> {dst_file}")
    shutil.copy2(source_path, dst_file)



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..inference_strategy import InferenceStrategy
from typing import Optional, Any


class A2uiTemplateManager(InferenceStrategy):

  def generate_system_prompt(
      self,
      role_description: str,
      workflow_description: str = "",
      ui_description: str = "",
      client_ui_capabilities: Optional[dict[str, Any]] = None,
      allowed_components: Optional[list[str]] = None,
      allowed_messages: Optional[list[str]] = None,
      include_schema: bool = False,
      include_examples: bool = False,
      validate_examples: bool = False,
  ) -> str:
    # TODO: Implementation logic for Template Manager
    raise NotImplementedError("This method is not yet implemented.")



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import logging
import re
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Set, Tuple, Union, Iterator

from jsonschema import Draft202012Validator

from .utils import wrap_as_json_array

if TYPE_CHECKING:
  from .catalog import A2uiCatalog

from .constants import (
    BASE_SCHEMA_URL,
    CATALOG_COMPONENTS_KEY,
    CATALOG_ID_KEY,
    CATALOG_STYLES_KEY,
    VERSION_0_8,
    VERSION_0_9,
)

# A2UI relaxed path pattern (extends RFC 6901 to support relative paths)
RELAXED_PATH_PATTERN = re.compile(
    r"^(?:(?:\/(?:[^~\/]|~[01])*)*|(?:[^~\/]|~[01])+(?:\/(?:[^~\/]|~[01])*)*)$"
)

# Recursion Limits
MAX_GLOBAL_DEPTH = 50
MAX_FUNC_CALL_DEPTH = 5

# Constants
COMPONENTS = "components"
ID = "id"
ROOT = "root"
PATH = "path"
FUNCTION_CALL = "functionCall"
CALL = "call"
ARGS = "args"


def _inject_additional_properties(
    schema: Dict[str, Any],
    source_properties: Dict[str, Any],
    mapping: Dict[str, str] = None,
) -> Tuple[Dict[str, Any], Set[str]]:
  """
  Recursively injects properties from source_properties into nodes with additionalProperties=True and sets additionalProperties=False.

  Args:
      schema: The target schema to traverse and patch.
      source_properties: A dictionary of top-level property groups (e.g., "components", "styles") from the source schema.

  Returns:
      A tuple containing:
      - The patched schema.
      - A set of keys from source_properties that were injected.
  """
  injected_keys = set()

  def recursive_inject(obj):
    if isinstance(obj, dict):
      new_obj = {}
      for k, v in obj.items():
        # If this node has additionalProperties=True, we inject the source properties
        if isinstance(v, dict) and v.get("additionalProperties") is True:
          if k in source_properties:
            injected_keys.add(k)
            new_node = dict(v)
            new_node["additionalProperties"] = False
            new_node["properties"] = {
                **new_node.get("properties", {}),
                **source_properties[k],
            }
            new_obj[k] = new_node
          else:  # No matching source group, keep as is but recurse children
            new_obj[k] = recursive_inject(v)
        else:  # Not a node with additionalProperties, recurse children
          new_obj[k] = recursive_inject(v)
      return new_obj
    elif isinstance(obj, list):
      return [recursive_inject(i) for i in obj]
    return obj

  return recursive_inject(schema), injected_keys


class A2uiValidator:
  """Validates the A2UI JSON payload against the provided schema and checks for integrity.

  Checks performed:
  1.  **JSON Schema Validation**: Ensures payload adheres to the A2UI schema.
  2.  **Component Integrity**:
      -   All component IDs are unique.
      -   A 'root' component exists.
      -   All unique component references point to valid IDs.
  3.  **Topology**:
      -   No circular references (including self-references).
      -   No orphaned components (all components must be reachable from 'root').
  4.  **Recursion Limits**:
      -   Global recursion depth limit (50).
      -   FunctionCall recursion depth limit (5).
  5.  **Path Syntax**:
      -   Validates JSON Pointer syntax for data paths.

  Args:
      a2ui_json: The JSON payload to validate.
      a2ui_schema: The schema to validate against.

  Raises:
      jsonschema.ValidationError: If the payload does not match the schema.
      ValueError: If integrity, topology, or recursion checks fail.
  """

  def __init__(self, catalog: "A2uiCatalog"):
    self._catalog = catalog
    self.version = getattr(catalog, "version", VERSION_0_8)
    self._validator = self._build_validator()

  def get_version(self) -> str:
    """Returns the A2UI protocol version."""
    return self.version

  def _build_validator(self) -> Draft202012Validator:
    """Builds a validator for the A2UI schema."""

    if self._catalog.version == VERSION_0_8:
      return self._build_0_8_validator()
    return self._build_0_9_validator()

  def _bundle_0_8_schemas(self) -> Dict[str, Any]:
    if not self._catalog.s2c_schema:
      return {}

    bundled = copy.deepcopy(self._catalog.s2c_schema)

    # Prepare catalog components and styles for injection
    source_properties = {}
    catalog_schema = self._catalog.catalog_schema
    if catalog_schema:
      if CATALOG_COMPONENTS_KEY in catalog_schema:
        # Special mapping for v0.8: "components" -> "component"
        source_properties["component"] = catalog_schema[CATALOG_COMPONENTS_KEY]
      if CATALOG_STYLES_KEY in catalog_schema:
        source_properties[CATALOG_STYLES_KEY] = catalog_schema[CATALOG_STYLES_KEY]

    bundled, _ = _inject_additional_properties(bundled, source_properties)
    return bundled

  def _build_0_8_validator(self) -> Draft202012Validator:
    """Builds a validator for the A2UI schema version 0.8."""
    bundled_schema = self._bundle_0_8_schemas()
    full_schema = wrap_as_json_array(bundled_schema)

    from referencing import Registry, Resource
    from referencing.jsonschema import DRAFT202012

    # Even in v0.8, we may have references to common_types.json or other files.
    base_uri = self._catalog.s2c_schema.get("$id", BASE_SCHEMA_URL)
    import os

    def get_sibling_uri(uri, filename):
      return os.path.join(os.path.dirname(uri), filename)

    common_types_uri = get_sibling_uri(base_uri, "common_types.json")

    resources = [
        (
            common_types_uri,
            Resource.from_contents(
                self._catalog.common_types_schema,
                default_specification=DRAFT202012,
            ),
        ),
        (
            "common_types.json",
            Resource.from_contents(
                self._catalog.common_types_schema,
                default_specification=DRAFT202012,
            ),
        ),
    ]

    registry = Registry().with_resources(resources)
    validator_schema = copy.deepcopy(full_schema)
    validator_schema["$schema"] = "https://json-schema.org/draft/2020-12/schema"

    return Draft202012Validator(validator_schema, registry=registry)

  def _build_0_9_validator(self) -> Draft202012Validator:
    """Builds a validator for the A2UI schema version 0.9+."""
    full_schema = wrap_as_json_array(self._catalog.s2c_schema)

    from referencing import Registry, Resource
    from referencing.jsonschema import DRAFT202012

    # v0.9 schemas (e.g. server_to_client.json) use relative references like
    # 'catalog.json#/$defs/anyComponent'. Since server_to_client.json has
    # $id: https://a2ui.org/specification/v0_9/server_to_client.json,
    # these resolve to https://a2ui.org/specification/v0_9/catalog.json.
    # We must register them using these absolute URIs.
    base_uri = self._catalog.s2c_schema.get("$id", BASE_SCHEMA_URL)
    import os

    def get_sibling_uri(uri, filename):
      return os.path.join(os.path.dirname(uri), filename)

    catalog_uri = get_sibling_uri(base_uri, "catalog.json")
    common_types_uri = get_sibling_uri(base_uri, "common_types.json")

    resources = [
        (
            common_types_uri,
            Resource.from_contents(
                self._catalog.common_types_schema,
                default_specification=DRAFT202012,
            ),
        ),
        (
            catalog_uri,
            Resource.from_contents(
                self._catalog.catalog_schema,
                default_specification=DRAFT202012,
            ),
        ),
        # Fallbacks for robustness
        (
            "catalog.json",
            Resource.from_contents(
                self._catalog.catalog_schema,
                default_specification=DRAFT202012,
            ),
        ),
        (
            "common_types.json",
            Resource.from_contents(
                self._catalog.common_types_schema,
                default_specification=DRAFT202012,
            ),
        ),
    ]
    # Also register the catalog ID if it's different from the catalog URI
    if self._catalog.catalog_id and self._catalog.catalog_id != catalog_uri:
      resources.append((
          self._catalog.catalog_id,
          Resource.from_contents(
              self._catalog.catalog_schema,
              default_specification=DRAFT202012,
          ),
      ))

    registry = Registry().with_resources(resources)
    validator_schema = copy.deepcopy(full_schema)
    validator_schema["$schema"] = "https://json-schema.org/draft/2020-12/schema"

    return Draft202012Validator(validator_schema, registry=registry)

  def validate(
      self,
      a2ui_json: Union[Dict[str, Any], List[Any]],
      root_id: Optional[str] = None,
      strict_integrity: bool = True,
  ) -> None:
    """Validates an A2UI messages against the schema."""
    messages = a2ui_json if isinstance(a2ui_json, list) else [a2ui_json]

    if self.version == VERSION_0_9:
      self._validate_0_9_custom(messages, root_id, strict_integrity)
    else:
      # Fallback to old behavior for v0.8
      errors = list(self._validator.iter_errors(messages))
      if errors:
        error = errors[0]
        msg = f"Validation failed: {error.message}"
        if error.context:
          msg += "\nContext failures:"
          for sub_error in error.context:
            msg += f"\n  - {sub_error.message}"
        raise ValueError(msg)

      for message in messages:
        if not isinstance(message, dict):
          continue

        components = None
        surface_id = None
        if "surfaceUpdate" in message:  # v0.8
          components = message["surfaceUpdate"].get(COMPONENTS)
          surface_id = message["surfaceUpdate"].get("surfaceId")

        if components:
          ref_map = extract_component_ref_fields(self._catalog)
          root_id = _find_root_id(messages, surface_id)
          _validate_component_integrity(
              root_id, components, ref_map, skip_root_check=not strict_integrity
          )
          analyze_topology(
              root_id, components, ref_map, raise_on_orphans=strict_integrity
          )

        _validate_recursion_and_paths(message)

  def _validate_0_9_custom(
      self,
      messages: List[Dict[str, Any]],
      root_id: Optional[str] = None,
      strict_integrity: bool = True,
  ) -> None:
    all_errors = []
    for idx, message in enumerate(messages):
      if not isinstance(message, dict):
        all_errors.append(f"messages[{idx}]: Is not an object")
        continue

      if "createSurface" in message:
        val = self._get_sub_validator("CreateSurfaceMessage")
        all_errors.extend(self._get_formatted_errors(val, message, f"messages[{idx}]"))
      elif "updateComponents" in message:
        all_errors.extend(
            self._get_update_components_errors(message, f"messages[{idx}]")
        )
      elif "updateDataModel" in message:
        val = self._get_sub_validator("UpdateDataModelMessage")
        all_errors.extend(self._get_formatted_errors(val, message, f"messages[{idx}]"))
      elif "deleteSurface" in message:
        val = self._get_sub_validator("DeleteSurfaceMessage")
        all_errors.extend(self._get_formatted_errors(val, message, f"messages[{idx}]"))
      else:
        keys = list(message.keys())
        all_errors.append(f"messages[{idx}]: Unknown message type with keys {keys}")

    if all_errors:
      msg = "Validation failed:\n" + "\n".join(f"  - {err}" for err in all_errors)
      raise ValueError(msg)

    # Integrity checks
    for message in messages:
      if not isinstance(message, dict):
        continue
      components = None
      surface_id = None
      if "updateComponents" in message and isinstance(
          message["updateComponents"], dict
      ):
        components = message["updateComponents"].get(COMPONENTS)
        surface_id = message["updateComponents"].get("surfaceId")

      if components:
        ref_map = extract_component_ref_fields(self._catalog)
        root_id = _find_root_id(messages, surface_id)
        _validate_component_integrity(
            root_id, components, ref_map, skip_root_check=not strict_integrity
        )
        analyze_topology(
            root_id, components, ref_map, raise_on_orphans=strict_integrity
        )

      _validate_recursion_and_paths(message)

  def _get_sub_validator(self, def_name: str) -> Draft202012Validator:
    sub_schema = self._catalog.s2c_schema.get("$defs", {}).get(def_name)
    if not sub_schema:
      raise ValueError(f"Definition {def_name} not found in schema")
    return Draft202012Validator(sub_schema, registry=self._validator._registry)

  def _get_formatted_errors(
      self, validator: Draft202012Validator, instance: Any, base_path: str
  ) -> List[str]:
    errors = list(validator.iter_errors(instance))
    formatted = []
    for err in errors:
      path_str = ".".join(str(p) for p in err.path)
      full_path = f"{base_path}.{path_str}" if path_str else base_path

      message = err.message
      if (
          (
              "Unevaluated properties are not allowed" in message
              or "Additional properties are not allowed" in message
          )
          and "(" in message
          and ")" in message
      ):
        message = message[message.find("(") + 1 : message.rfind(")")]

      formatted.append(f"{full_path}: {message}")
    return formatted

  def _get_update_components_errors(
      self, message: Dict[str, Any], path: str
  ) -> List[str]:
    errors = []
    if "version" not in message or message["version"] != "v0.9":
      errors.append(f"{path}: Invalid version, expected 'v0.9'")

    uc = message.get("updateComponents")
    if not isinstance(uc, dict):
      errors.append(f"{path}: Expected updateComponents to be an object")
      return errors

    if "surfaceId" not in uc or not isinstance(uc["surfaceId"], str):
      errors.append(f"{path}.updateComponents: Invalid or missing surfaceId")

    components = uc.get("components")
    if not isinstance(components, list):
      errors.append(f"{path}.updateComponents: Expected components to be an array")
      return errors

    for idx, comp in enumerate(components):
      comp_id = comp.get("id")
      comp_path = (
          f"{path}.updateComponents.components[id='{comp_id}']"
          if comp_id
          else f"{path}.updateComponents.components[{idx}]"
      )
      errors.extend(self._get_single_component_errors(comp, comp_path))

    return errors

  def _get_single_component_errors(self, comp: Dict[str, Any], path: str) -> List[str]:
    if not isinstance(comp, dict):
      return [f"{path}: Component is not an object"]

    comp_type = comp.get("component")
    if not comp_type:
      return [f"{path}: Missing 'component' field"]

    catalog = self._catalog.catalog_schema
    if not catalog or "components" not in catalog:
      return [f"{path}: Catalog schema or components missing"]

    comp_schema = catalog["components"].get(comp_type)
    if not comp_schema:
      return [f"{path}: Unknown component: {comp_type}"]

    temp_schema = {
        "$schema": "https://json-schema.org/draft/2020-12/schema",
        "$ref": f"catalog.json#/components/{comp_type}",
    }

    validator = Draft202012Validator(temp_schema, registry=self._validator._registry)
    return self._get_formatted_errors(validator, comp, path)


def _find_root_id(
    messages: List[Dict[str, Any]], surface_id: Optional[str] = None
) -> Optional[str]:
  """
  Finds the root id from a list of A2UI messages for a given surface.
  - For v0.8, the root id is in the beginRendering message.
  - For v0.9+, the root id is 'root'.
  """
  for message in messages:
    if not isinstance(message, dict):
      continue
    if "beginRendering" in message:
      if surface_id and message["beginRendering"].get("surfaceId") != surface_id:
        continue
      return message["beginRendering"].get(ROOT, ROOT)
    if "createSurface" in message:
      if surface_id and message["createSurface"].get("surfaceId") != surface_id:
        continue
      return ROOT
  return None


def _validate_component_integrity(
    root_id: Optional[str],
    components: List[Dict[str, Any]],
    ref_fields_map: Dict[str, tuple[Set[str], Set[str]]],
    skip_root_check: bool = False,
) -> None:
  """
  Validates that:
  1. All component IDs are unique.
  2. A 'root' component exists.
  3. All references point to existing IDs.
  """
  ids: Set[str] = set()

  # 1. Collect IDs and check for duplicates
  for comp in components:
    comp_id = comp.get(ID)
    if comp_id is None:
      continue

    if comp_id in ids:
      raise ValueError(f"Duplicate component ID: {comp_id}")
    ids.add(comp_id)

  # 2. Check for root component
  if not skip_root_check and root_id is not None and root_id not in ids:
    raise ValueError(f"Missing root component: No component has id='{root_id}'")

  # 3. Check for dangling references using helper
  # In an incremental update (root_id is None), components may reference IDs already on the client.
  if root_id is not None and not skip_root_check:
    for comp in components:
      for ref_id, field_name in get_component_references(comp, ref_fields_map):
        if ref_id not in ids:
          raise ValueError(
              f"Component '{comp.get(ID)}' references non-existent component '{ref_id}'"
              f" in field '{field_name}'"
          )


def analyze_topology(
    root_id: Optional[str],
    components: List[Dict[str, Any]],
    ref_fields_map: Dict[str, tuple[Set[str], Set[str]]],
    raise_on_orphans: bool = False,
) -> Set[str]:
  """
  Analyzes the topology of the component tree and returns reachable component IDs.

  Args:
      root_id: The ID of the root component.
      components: The list of components.
      ref_fields_map: Map of component reference fields.
      raise_on_orphans: If True, raises ValueError if any components are unreachable from root.

  Returns:
      A set of reachable component IDs.

  Raises:
      ValueError: On circular references or self-references.
  """
  adj_list: Dict[str, List[str]] = {}
  all_ids: Set[str] = set()

  # Build Adjacency List
  for comp in components:
    comp_id = comp.get(ID)
    if comp_id is None:
      continue

    all_ids.add(comp_id)
    if comp_id not in adj_list:
      adj_list[comp_id] = []

    for ref_id, field_name in get_component_references(comp, ref_fields_map):
      if ref_id == comp_id:
        raise ValueError(
            f"Self-reference detected: Component '{comp_id}' references itself in field"
            f" '{field_name}'"
        )
      adj_list[comp_id].append(ref_id)

  # Detect Cycles and Depth using DFS
  visited: Set[str] = set()
  recursion_stack: Set[str] = set()

  def dfs(node_id: str, depth: int):
    if depth > MAX_GLOBAL_DEPTH:
      raise ValueError(
          f"Global recursion limit exceeded: logical depth > {MAX_GLOBAL_DEPTH}"
      )

    visited.add(node_id)
    recursion_stack.add(node_id)

    for neighbor in adj_list.get(node_id, []):
      if neighbor not in visited:
        dfs(neighbor, depth + 1)
      elif neighbor in recursion_stack:
        raise ValueError(
            f"Circular reference detected involving component '{neighbor}'"
        )

    recursion_stack.remove(node_id)

  if root_id is not None:
    if root_id in all_ids:
      dfs(root_id, 0)

    # Check for Orphans if requested
    if raise_on_orphans:
      orphans = all_ids - visited
      if orphans:
        sorted_orphans = sorted(list(orphans))
        raise ValueError(
            f"Component '{sorted_orphans[0]}' is not reachable from '{root_id}'"
        )
  else:
    # No root provided (e.g. partial update): we traverse everything to check for cycles
    for node_id in sorted(list(all_ids)):
      if node_id not in visited:
        dfs(node_id, 0)

  return visited


def extract_component_required_fields(
    catalog: "A2uiCatalog",
) -> Dict[str, Set[str]]:
  """
  Parses the catalog/schema to identify which component properties are required.
  Returns a map: { component_name: set_of_required_fields }
  """
  req_map = {}

  all_components = {}
  # Version aware extraction
  if catalog.version == VERSION_0_8:
    # Search for components in s2c schema properties
    try:
      s2c = catalog.s2c_schema or {}
      props = s2c.get("properties", {})
      if "surfaceUpdate" in props:
        su = props["surfaceUpdate"].get("properties", {})
        if "components" in su:
          items = su["components"].get("items", {})
          if "properties" in items:
            comp_wrapper = items["properties"].get("component", {})
            all_components = comp_wrapper.get("properties", {})
    except Exception:
      pass

    if not all_components and catalog.catalog_schema:
      all_components = catalog.catalog_schema.get(COMPONENTS, {})
  else:  # v0.9+
    all_components = catalog.catalog_schema.get(COMPONENTS, {})

  for comp_name, comp_schema in all_components.items():
    required_fields = set()

    def extract_from_props(cs: Dict[str, Any]):
      if not isinstance(cs, dict):
        return

      if "required" in cs and isinstance(cs["required"], list):
        required_fields.update(req for req in cs["required"] if req != "component")

      # Recurse into allOf/oneOf/anyOf
      for key in ["allOf", "oneOf", "anyOf"]:
        if key in cs:
          for sub in cs[key]:
            extract_from_props(sub)

    extract_from_props(comp_schema)

    if required_fields:
      req_map[comp_name] = required_fields

  return req_map


def extract_component_ref_fields(
    catalog: "A2uiCatalog",
) -> Dict[str, tuple[Set[str], Set[str]]]:
  """
  Parses the catalog/schema to identify which component properties reference other components.
  Returns a map: { component_name: (set_of_single_ref_fields, set_of_list_ref_fields) }
  """
  ref_map = {}

  all_components = {}
  # Version aware extraction
  if catalog.version == VERSION_0_8:
    # Search for components in s2c schema properties
    try:
      # Try nested path: surfaceUpdate -> components -> items -> properties -> component -> properties
      s2c = catalog.s2c_schema or {}
      props = s2c.get("properties", {})

      # Might be in surfaceUpdate or beginRendering component definitions
      if "surfaceUpdate" in props:
        su = props["surfaceUpdate"].get("properties", {})
        if "components" in su:
          items = su["components"].get("items", {})
          if "properties" in items:
            comp_wrapper = items["properties"].get("component", {})
            all_components = comp_wrapper.get("properties", {})
    except Exception:
      logging.warning("Failed to extract component ref fields from v0.8 schema")

    # Also check catalog schema if available
    if not all_components and catalog.catalog_schema:
      all_components = catalog.catalog_schema.get(COMPONENTS, {})
  else:  # v0.9+
    # In v0.9, components are defined in the catalog itself
    all_components = catalog.catalog_schema.get(COMPONENTS, {})

  # Helper to check if a property schema looks like a ComponentId reference
  def is_component_id_ref(prop_schema: Dict[str, Any]) -> bool:
    if not isinstance(prop_schema, dict):
      return False
    ref = prop_schema.get("$ref", "")
    if isinstance(ref, str) and (
        ref.endswith("ComponentId") or ref.endswith("child") or "/child" in ref
    ):
      return True

    # Inline check
    if (
        prop_schema.get("type") == "string"
        and prop_schema.get("title") == "ComponentId"
    ):
      return True

    # Check oneOf/anyOf for refs
    for key in ["oneOf", "anyOf", "allOf"]:
      if key in prop_schema:
        for sub in prop_schema[key]:
          if is_component_id_ref(sub):
            return True
    return False

  def is_child_list_ref(prop_schema: Dict[str, Any]) -> bool:
    if not isinstance(prop_schema, dict):
      return False
    ref = prop_schema.get("$ref", "")
    if isinstance(ref, str) and (
        ref.endswith("ChildList") or ref.endswith("children") or "/children" in ref
    ):
      return True

    # Inline check
    if prop_schema.get("type") == "object":
      props = prop_schema.get("properties", {})
      if "explicitList" in props or "template" in props or "componentId" in props:
        return True

    # Or array of ComponentIds
    if prop_schema.get("type") == "array":
      items = prop_schema.get("items", {})
      if is_component_id_ref(items):
        return True

    # Check oneOf/anyOf for refs
    for key in ["oneOf", "anyOf", "allOf"]:
      if key in prop_schema:
        for sub in prop_schema[key]:
          if is_child_list_ref(sub):
            return True
    return False

  for comp_name, comp_schema in all_components.items():
    single_refs = set()
    list_refs = set()

    def extract_from_props(cs: Dict[str, Any]):
      if not isinstance(cs, dict):
        return
      props = cs.get("properties", {})
      for prop_name, prop_schema in props.items():
        if is_component_id_ref(prop_schema) or prop_name in [
            "child",
            "contentChild",
            "entryPointChild",
        ]:
          single_refs.add(prop_name)
        elif is_child_list_ref(prop_schema) or prop_name == "children":
          list_refs.add(prop_name)

      # Recurse into allOf/oneOf for properties
      for key in ["allOf", "oneOf", "anyOf"]:
        if key in cs:
          for sub in cs[key]:
            extract_from_props(sub)

    extract_from_props(comp_schema)

    if single_refs or list_refs:
      ref_map[comp_name] = (single_refs, list_refs)

  return ref_map


def get_component_references(
    component: Dict[str, Any], ref_fields_map: Dict[str, tuple[Set[str], Set[str]]]
) -> Iterator[Tuple[str, str]]:
  """
  Helper to extract all referenced component IDs from a component.
  Yields (referenced_id, field_name).
  """
  # Support both v0.8 and v0.9+
  comp_val = component.get("component")
  if isinstance(comp_val, str):
    # v0.9 flattened
    yield from get_refs_recursively(comp_val, component, ref_fields_map)
  elif isinstance(comp_val, dict):
    # v0.8 structured
    for c_type, c_props in comp_val.items():
      # Recurse into the properties container
      if isinstance(c_props, dict):
        yield from get_refs_recursively(c_type, c_props, ref_fields_map)


def get_refs_recursively(
    comp_type: str,
    props: Dict[str, Any],
    ref_fields_map: Dict[str, tuple[Set[str], Set[str]]],
) -> Iterator[Tuple[str, str]]:
  if not comp_type or not isinstance(props, dict):
    return

  single_refs, list_refs = ref_fields_map.get(comp_type, (set(), set()))

  # Standard A2UI reference fields to check as heuristics if not explicitly mapped
  HEURISTIC_SINGLE = {
      "child",
      "contentChild",
      "entryPointChild",
      "detail",
      "summary",
      "root",
  }
  HEURISTIC_LIST = {"children", "explicitList", "template"}

  for key, value in props.items():
    is_ref = False
    if key in single_refs or key in HEURISTIC_SINGLE:
      if isinstance(value, str):
        yield value, key
        is_ref = True
      elif isinstance(value, dict) and "componentId" in value:  # ChildList template
        yield value["componentId"], f"{key}.componentId"
        is_ref = True
    elif key in list_refs or key in HEURISTIC_LIST:
      if isinstance(value, list):
        for item in value:
          if isinstance(item, str):
            yield item, key
            is_ref = True
      elif isinstance(value, dict):
        if "explicitList" in value:
          for item in value["explicitList"]:
            if isinstance(item, str):
              yield item, f"{key}.explicitList"
              is_ref = True
        elif "template" in value:
          template = value["template"]
          if isinstance(template, dict) and "componentId" in template:
            yield template["componentId"], f"{key}.template.componentId"
            is_ref = True
        elif "componentId" in value:
          yield value["componentId"], f"{key}.componentId"
          is_ref = True

    # Special handling for 'tabs' or other nested arrays
    if isinstance(value, list) and key not in list_refs:
      for idx, item in enumerate(value):
        if isinstance(item, dict):
          # Check for common patterns like {title, child}
          child_id = item.get("child")
          if child_id and isinstance(child_id, str):
            yield child_id, f"{key}[{idx}].child"


def _validate_recursion_and_paths(data: Any) -> None:
  """
  Validates:
  1. Global recursion depth limit (50).
  2. FunctionCall recursion depth limit (5).
  3. Path syntax for DataBindings/DataModelUpdates.
  """

  def traverse(item: Any, global_depth: int, func_depth: int):
    if global_depth > MAX_GLOBAL_DEPTH:
      raise ValueError(f"Global recursion limit exceeded: Depth > {MAX_GLOBAL_DEPTH}")

    if isinstance(item, list):
      for x in item:
        traverse(x, global_depth + 1, func_depth)
      return

    if isinstance(item, dict):
      # Check for path
      if PATH in item and isinstance(item[PATH], str):
        path = item[PATH]
        if not re.fullmatch(RELAXED_PATH_PATTERN, path):
          raise ValueError(f"Invalid path syntax: '{path}'")

      # Check for FunctionCall
      is_func = CALL in item and ARGS in item

      if is_func:
        if func_depth >= MAX_FUNC_CALL_DEPTH:
          raise ValueError(
              f"Recursion limit exceeded: {FUNCTION_CALL} depth > {MAX_FUNC_CALL_DEPTH}"
          )

        # Increment func_depth only for 'args', but global_depth matches traversal
        for k, v in item.items():
          if k == ARGS:
            traverse(v, global_depth + 1, func_depth + 1)
          else:
            traverse(v, global_depth + 1, func_depth)
      else:
        for v in item.values():
          traverse(v, global_depth + 1, func_depth)

  traverse(data, 0, 0)



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import logging
import re
from typing import Any, Dict, List


logger = logging.getLogger(__name__)


def parse_and_fix(payload: str) -> List[Dict[str, Any]]:
  """Validates and applies autofixes to a raw JSON string and returns the parsed payload.

  Args:
    payload: The raw JSON string from the LLM.

  Returns:
    A parsed and potentially fixed payload (list of dicts).
  """
  normalized_payload = _normalize_smart_quotes(payload)
  try:
    a2ui_json = _parse(normalized_payload)
    return a2ui_json
  except (
      json.JSONDecodeError,
      ValueError,
  ) as e:
    logger.warning(f"Initial A2UI payload validation failed: {e}")
    updated_payload = _remove_trailing_commas(normalized_payload)
    a2ui_json = _parse(updated_payload)
    return a2ui_json


def _parse(payload: str) -> List[Dict[str, Any]]:
  """Parses the payload and returns a list of A2UI JSON objects."""
  try:
    a2ui_json = json.loads(payload)
    if not isinstance(a2ui_json, list):
      logger.info("Received a single JSON object, wrapping in a list for validation.")
      a2ui_json = [a2ui_json]
    return a2ui_json
  except json.JSONDecodeError as e:
    logger.error(f"Failed to parse JSON: {e}")
    raise ValueError(f"Failed to parse JSON: {e}")


def _normalize_smart_quotes(json_str: str) -> str:
  """Replaces smart (curly) quotes with standard straight quotes."""
  return (
      json_str.replace("\u201C", '"')
      .replace("\u201D", '"')
      .replace("\u2018", "'")
      .replace("\u2019", "'")
  )


def _remove_trailing_commas(json_str: str) -> str:
  """Attempts to remove trailing commas from a JSON string.

  Args:
    json_str: The raw JSON string from the LLM.

  Returns:
    A potentially fixed JSON string.
  """
  # Fix trailing commas: identifying commas followed by optional whitespace and a closing bracket (]) or brace (}).
  fixed_json = re.sub(r",(?=\s*[\]}])", "", json_str)

  if fixed_json != json_str:
    logger.warning("Detected trailing commas in LLM output; applied autofix.")

  return fixed_json



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def remove_strict_validation(schema):
  if isinstance(schema, dict):
    new_schema = {k: remove_strict_validation(v) for k, v in schema.items()}
    if (
        'additionalProperties' in new_schema
        and new_schema['additionalProperties'] is False
    ):
      del new_schema['additionalProperties']
    return new_schema
  elif isinstance(schema, list):
    return [remove_strict_validation(item) for item in schema]
  return schema



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from typing import Any, Optional


@dataclass
class ResponsePart:
  """Represents a part of the LLM response.

  Attributes:
      text: The conversational text part. Can be an empty string.
      a2ui_json: The parsed A2UI JSON data, always a list of dictionaries if
        it contains A2UI messages. None if this part only contains trailing
        text.
  """

  text: str = ""
  a2ui_json: Optional[Any] = None



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import pytest
from typing import Any, Dict, List
from a2ui.schema.catalog import A2uiCatalog
from a2ui.schema.constants import (
    A2UI_SCHEMA_BLOCK_START,
    A2UI_SCHEMA_BLOCK_END,
    VERSION_0_8,
    VERSION_0_9,
)
from a2ui.basic_catalog.constants import BASIC_CATALOG_NAME


def test_catalog_id_property():
  catalog_id = "https://a2ui.org/specification/v0_9/catalogs/basic/catalog.json"
  catalog = A2uiCatalog(
      version=VERSION_0_8,
      name=BASIC_CATALOG_NAME,
      s2c_schema={},
      common_types_schema={},
      catalog_schema={"catalogId": catalog_id},
  )
  assert catalog.catalog_id == catalog_id


def test_catalog_id_missing_raises_error():
  catalog = A2uiCatalog(
      version=VERSION_0_8,
      name=BASIC_CATALOG_NAME,
      s2c_schema={},
      common_types_schema={},
      catalog_schema={},  # No catalogId
  )
  with pytest.raises(
      ValueError, match=f"Catalog '{BASIC_CATALOG_NAME}' missing catalogId"
  ):
    _ = catalog.catalog_id


def test_resolve_examples_path_handling():
  from a2ui.schema.catalog import resolve_examples_path

  assert resolve_examples_path(None) is None
  assert resolve_examples_path("/absolute/examples") == "/absolute/examples"
  assert resolve_examples_path("file:///absolute/examples") == "/absolute/examples"

  with pytest.raises(ValueError, match="Unsupported examples URL scheme"):
    resolve_examples_path("https://a2ui.org/examples")


def test_catalog_config_from_path_schemes():
  from a2ui.schema.catalog import CatalogConfig
  # Test local path
  config = CatalogConfig.from_path(
      name="test_file", catalog_path="relative_path/to/catalog.json"
  )
  assert config.provider.path == "relative_path/to/catalog.json"

  # Test file:// scheme
  config = CatalogConfig.from_path(
      name="test_file", catalog_path="file:///absolute_path/to/catalog.json"
  )
  assert config.provider.path == "/absolute_path/to/catalog.json"

  # Test HTTP raises NotImplementedError
  with pytest.raises(NotImplementedError, match="HTTP support is coming soon."):
    CatalogConfig.from_path(
        name="test_http", catalog_path="http://a2ui.org/catalog.json"
    )

  # Test unsupported scheme raises ValueError
  with pytest.raises(ValueError, match="Unsupported catalog URL scheme"):
    CatalogConfig.from_path(name="test_ftp", catalog_path="ftp://a2ui.org/catalog.json")


def test_basic_catalog_get_config_examples_path():
  from a2ui.basic_catalog.provider import BasicCatalog
  from a2ui.schema.constants import VERSION_0_9

  # Test get_config with file:// scheme examples path
  config = BasicCatalog.get_config(
      version=VERSION_0_9, examples_path="file:///absolute/examples"
  )
  assert config.examples_path == "/absolute/examples"



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module for the A2UI Part Converter.

This module provides the `A2uiPartConverter` which acts as a catalog-aware GenAI to A2A
part converter. It handles both tool-based A2UI (via the `send_a2ui_json_to_client` tool response)
and text-based A2UI (extracted and healed via A2UI custom tags), validating the structures
against the active A2UI catalog schema.

Key Components:
  * `A2uiPartConverter`: A catalog-aware GenAI to A2A part converter.

Usage Example:

  Typically used internally by the A2uiEventConverter or manually configured for custom part conversions:

  ```python
  converter = A2uiPartConverter(a2ui_catalog=my_catalog)
  a2a_parts = converter.convert(genai_part)
  ```
"""

import logging
from typing import Optional


from a2a import types as a2a_types
from a2ui.a2a.parts import create_a2ui_part, parse_response_to_parts
from a2ui.parser.parser import has_a2ui_parts
from a2ui.schema import constants
from a2ui.schema.catalog import A2uiCatalog
from google.adk.a2a.converters import part_converter
from google.adk.utils.feature_decorator import experimental
from google.genai import types as genai_types

logger = logging.getLogger(__name__)


@experimental
class A2uiPartConverter:
  """A catalog-aware GenAI to A2A part converter.

  This converter handles both tool-based A2UI (via `send_a2ui_json_to_client`)
  and text-based A2UI (via A2UI delimiter tags). It uses the provided
  catalog to validate and fix JSON payloads.
  """

  def __init__(
      self,
      a2ui_catalog: A2uiCatalog,
      bypass_tool_check: bool = False,
      fallback_text: Optional[str] = None,
  ):
    self._catalog = a2ui_catalog
    self._bypass_tool_check = bypass_tool_check
    self._fallback_text = fallback_text

  def convert(self, part: genai_types.Part) -> list[a2a_types.Part]:
    """Converts a GenAI part to A2A parts, with A2UI validation.

    Args:
        part: The GenAI part to convert.

    Returns:
        A list of A2A parts.
    """
    # 1. Handle Tool Responses (FunctionResponse)
    if function_response := part.function_response:
      is_send_a2ui_json_to_client_response = (
          function_response.name == constants.A2UI_TOOL_NAME
      )

      if is_send_a2ui_json_to_client_response or self._bypass_tool_check:
        response_dict = function_response.response or {}

        if constants.A2UI_TOOL_ERROR_KEY in response_dict:
          logger.warning(
              f"A2UI tool call failed: {response_dict[constants.A2UI_TOOL_ERROR_KEY]}"
          )
          return []

        if (
            isinstance(response_dict, dict)
            and constants.A2UI_VALIDATED_JSON_KEY in response_dict
        ):
          json_data = response_dict.get(constants.A2UI_VALIDATED_JSON_KEY)
          if json_data:
            return [create_a2ui_part(message) for message in json_data]

        if is_send_a2ui_json_to_client_response:
          logger.info("No result in A2UI tool response")
          return []

      # Handle generic/other tool responses that returned a string containing A2UI tags.
      if function_response.response and function_response.response.get("result"):
        result = function_response.response.get("result")
        if has_a2ui_parts(result):
          return parse_response_to_parts(
              result,
              validator=self._catalog.validator,
              fallback_text=self._fallback_text,
          )

    # 2. Handle Tool Calls (FunctionCall) - Skip sending to client
    if (
        function_call := part.function_call
    ) and function_call.name == constants.A2UI_TOOL_NAME:
      return []

    # 3. Handle Text-based A2UI (TextPart)
    if text := part.text:
      if has_a2ui_parts(text):
        return parse_response_to_parts(
            text,
            validator=self._catalog.validator,
            fallback_text=self._fallback_text,
        )

    # 4. Default conversion for other parts
    converted_part = part_converter.convert_genai_part_to_a2a_part(part)
    return [converted_part] if converted_part else []



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from typing import List, Optional, Any
from .response_part import ResponsePart
from ..schema.constants import A2UI_OPEN_TAG, A2UI_CLOSE_TAG
from .payload_fixer import parse_and_fix


_A2UI_BLOCK_PATTERN = re.compile(
    f"{re.escape(A2UI_OPEN_TAG)}(.*?){re.escape(A2UI_CLOSE_TAG)}", re.DOTALL
)


def has_a2ui_parts(content: str) -> bool:
  """Checks if the content has A2UI parts."""
  return A2UI_OPEN_TAG in content and A2UI_CLOSE_TAG in content


def _sanitize_json_string(json_string: str) -> str:
  """Sanitizes the JSON string by removing markdown code blocks."""
  json_string = json_string.strip()
  if json_string.startswith("```json"):
    json_string = json_string[len("```json") :]
  elif json_string.startswith("```"):
    json_string = json_string[len("```") :]
  if json_string.endswith("```"):
    json_string = json_string[: -len("```")]
  json_string = json_string.strip()
  return json_string


def parse_response(content: str) -> List[ResponsePart]:
  """
  Parses the LLM response into a list of ResponsePart objects.

  Args:
      content: The raw LLM response.

  Returns:
      A list of ResponsePart objects.

  Raises:
      ValueError: If no A2UI tags are found or if the JSON part is invalid.
  """
  matches = list(_A2UI_BLOCK_PATTERN.finditer(content))

  if not matches:
    raise ValueError(
        f"A2UI tags '{A2UI_OPEN_TAG}' and '{A2UI_CLOSE_TAG}' not found in response."
    )

  response_parts = []
  last_end = 0

  for match in matches:
    start, end = match.span()
    # Text preceding the JSON block
    text_part = content[last_end:start].strip()

    # The JSON content within the tags
    json_string = match.group(1)
    json_string_cleaned = _sanitize_json_string(json_string)
    if not json_string_cleaned:
      raise ValueError("A2UI JSON part is empty.")

    json_data = parse_and_fix(json_string_cleaned)
    response_parts.append(ResponsePart(text=text_part, a2ui_json=json_data))
    last_end = end

  # Trailing text after the last JSON block
  trailing_text = content[last_end:].strip()
  if trailing_text:
    response_parts.append(ResponsePart(text=trailing_text, a2ui_json=None))

  return response_parts
